#!/usr/bin/env python3

import sys
import os
import argparse
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from lottie.exporters.tgs_validator import TgsValidator, Severity
from lottie import __version__


parser = argparse.ArgumentParser(
    description="Checks a lottie or tgs file to see if it's compatible with telegram stickers",
    formatter_class=argparse.RawTextHelpFormatter
)
parser.add_argument("--version", "-v", action="version", version="%(prog)s - python-lottie " + __version__)
parser.add_argument(
    "infile",
    help="Input file"
)
parser.add_argument(
    "--level", "-l",
    choices=list(Severity.__members__.keys()),
    help="Error level:\n"
    "* Note   : the feature is not officially supported but works regardless\n"
    "* Warning: the feature is not supported, might result in different animations than expected\n"
    "* Error  : Telegram will not recognize the sticker\n",
    default="Note",
)


if __name__ == "__main__":
    ns = parser.parse_args()
    severity = Severity.__members__[ns.level]

    validator = TgsValidator(severity)
    validator.check_file(ns.infile)
    if validator.errors:
        sys.stdout.write("\n".join(map(str, validator.errors))+"\n")
        sys.exit(1)



#!/usr/bin/env python3

import json
import sys
import os
import shutil
import argparse
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from lottie.utils.linediff import difflines
from lottie.objects import Animation
from lottie.parsers.tgs import parse_tgs
from lottie import __version__


parser = argparse.ArgumentParser(
    description="Shows a side-by-side diff of the human-readable rendition of two tgs / lottie files",
)
parser.add_argument("--version", "-v", action="version", version="%(prog)s - python-lottie " + __version__)
parser.add_argument(
    "file1",
    help="Left file"
)
parser.add_argument(
    "file2",
    help="Right file"
)

if __name__ == "__main__":
    ns = parser.parse_args()
    width = shutil.get_terminal_size((-1, -1)).columns
    if width < 10:
        width = None
    else:
        width = int((width - 3) / 2)
    a1 = parse_tgs(ns.file1)
    a2 = parse_tgs(ns.file2)
    difflines(a1, a2, width, width)



from xml.dom import minidom
import copy
from uuid import uuid4

from .utils import *
from lottie.parsers.sif.sif.core import ObjectRegistry


class XmlDescriptor:
    def __init__(self, name):
        self.name = name
        self.att_name = name.replace("-", "_").replace("[", "_").replace("]", "").replace(".", "_")

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        raise NotImplementedError

    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        raise NotImplementedError

    def from_python(self, value):
        raise NotImplementedError

    def initialize_object(self, dict, obj):
        if self.att_name in dict:
            setattr(obj, self.att_name, self.from_python(dict[self.att_name]))
        else:
            setattr(obj, self.att_name, self.default())

    def clean(self, value):
        return self.from_python(value)

    def default(self):
        return None

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.name)


class TypedXmlDescriptor(XmlDescriptor):
    def __init__(self, name, type=str, default_value=None, att_name=None):
        super().__init__(name)
        self.type = type
        self.default_value = default_value
        if att_name is not None:
            self.att_name = att_name

    def from_python(self, value):
        if value is None and self.default_value is None:
            return None
        if not value_isinstance(value, self.type):
            return self.type(value)
        return value

    def default(self):
        return copy.deepcopy(self.default_value)


class ValueReference:
    def __init__(self, id, value=None):
        self.value = value
        self.id = id

    @classmethod
    def from_registry(cls, id, registry: ObjectRegistry):
        return cls(id, registry.get_object(id))


class XmlAttribute(TypedXmlDescriptor):
    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        xml_str = parent.getAttribute(self.name)
        if xml_str:
            if xml_str.startswith(":") and xml_str[1:] in registry.registry:
                value = ValueReference.from_registry(xml_str[1:], registry)
            else:
                value = value_from_xml_string(xml_str, self.type, registry)
            setattr(obj, self.att_name, value)

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        value = getattr(obj, self.att_name)
        if value is not None:
            if isinstance(value, ValueReference):
                xml_str = ":" + value.id
            else:
                xml_str = value_to_xml_string(value, self.type)
            parent.setAttribute(self.name, xml_str)


class XmlFixedAttribute(XmlDescriptor):
    def __init__(self, name, value, type=str):
        super().__init__(name)
        self.value = value
        self.type = type

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        parent.setAttribute(self.name, value_to_xml_string(self.value, self.type))

    def from_python(self, value):
        if value != self.value:
            raise ValueError("Value of %s should be %s, got %s" % (self.name, self.value, value))
        return value

    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        xml_str = parent.getAttribute(self.name)
        setattr(obj, self.att_name, value_from_xml_string(xml_str, self.type, registry))

    def default(self):
        return self.value


class XmlSimpleElement(TypedXmlDescriptor):
    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        cn = xml_first_element_child(parent, self.name, allow_none=True)
        if cn:
            value = value_from_xml_string(xml_text(cn), self.type, registry)
        else:
            value = self.default_value

        setattr(obj, self.att_name, value)

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        value = getattr(obj, self.att_name)
        if value is not None:
            parent.appendChild(xml_make_text(dom, self.name, value_to_xml_string(value, self.type)))


class XmlMeta(TypedXmlDescriptor):
    def from_xml(self, obj, parent: minidom.Element, registry: ObjectRegistry):
        for cn in xml_child_elements(parent, "meta"):
            if cn.getAttribute("name") == self.name:
                value = value_from_xml_string(cn.getAttribute("content"), self.type, registry)
                break
        else:
            value = self.default_value

        setattr(obj, self.att_name, value)

    def to_xml(self, obj, parent: minidom.Element, dom: minidom.Document):
        value = getattr(obj, self.att_name)
        if value is not None:
            meta = parent.appendChild(dom.createElement("meta"))
            meta.setAttribute("name", self.name)
            meta.setAttribute("content", value_to_xml_string(value, self.type))



import sys
import json
import gzip
import codecs

from .base import exporter
from ..utils.file import open_file
from ..parsers.baseporter import ExtraOption
from .tgs_validator import TgsValidator


@exporter("Lottie JSON", ["json"], [], {"pretty"}, "lottie")
def export_lottie(animation, file, pretty=False):
    with open_file(file) as fp:
        kw = {}
        if pretty:
            kw = dict(indent=4)
        json.dump(animation.to_dict(), fp, **kw)


@exporter("Telegram Animated Sticker", ["tgs"], [
    ExtraOption("no_sanitize", help="Disable Sticker fit", action="store_false", dest="sanitize"),
    ExtraOption("no_validate", help="Disable feature validation", action="store_false", dest="validate"),
])
def export_tgs(animation, file, sanitize=False, validate=False):
    if sanitize:
        animation.tgs_sanitize()

    with gzip.open(file, "wb") as gzfile:
        lottie_dict = animation.to_dict()
        lottie_dict["tgs"] = 1
        json.dump(lottie_dict, codecs.getwriter('utf-8')(gzfile))

    if validate:
        validator = TgsValidator()
        validator(animation)
        validator.check_file_size(file)
        if validator.errors:
            sys.stdout.write("\n".join(map(str, validator.errors))+"\n")


class HtmlOutput:
    def __init__(self, animation, file):
        self.animation = animation
        self.file = file

    def style(self):
        self.file.write("""
    <style>
        #bodymovin { width: %spx; height: %spx; margin: auto;
            background-color: white;
            background-size: 64px 64px;
            background-image:
                linear-gradient(to right, rgba(0, 0, 0, .3) 50%%, transparent 50%%),
                linear-gradient(to bottom, rgba(0, 0, 0, .3) 50%%, transparent 50%%),
                linear-gradient(to bottom, white 50%%, transparent 50%%),
                linear-gradient(to right, transparent 50%%, rgba(0, 0, 0, .5) 50%%);
        }
    </style>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/bodymovin/5.5.3/lottie.js"></script>
    """ % (self.animation.width, self.animation.height))

    def body_pre(self):
        self.file.write("""
<div id="bodymovin"></div>

<script>
    var animData = {
        container: document.getElementById('bodymovin'),
        renderer: 'svg',
        loop: true,
        autoplay: true,
        """)

    def body_embedded(self):
        self.file.write("animationData: ")
        export_lottie(self.animation, self.file, True)

    def body_post(self):
        self.file.write("""
    };
    var anim = bodymovin.loadAnimation(animData);
</script>""")

    def html_begin(self):
        self.file.write("""<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8" />
    <style>
        html, body { width: 100%; height: 100%; margin: 0; }
        body { display: flex; }
    </style>""")
        self.style()
        self.file.write("</head><body>")

    def html_end(self):
        self.file.write("</body></html>")


@exporter("Lottie HTML", ["html", "htm"])
def export_embedded_html(animation, file):
    with open_file(file) as fp:
        out = HtmlOutput(animation, fp)
        out.html_begin()
        out.body_pre()
        out.body_embedded()
        out.body_post()
        out.html_end()


def export_linked_html(animation, file, path):
    with open_file(file) as fp:
        out = HtmlOutput(animation, fp)
        out.html_begin()
        out.body_pre()
        file.write("path: %r" % path)
        out.body_post()
        out.html_end()



import io
import json
import gzip
from ..objects import Animation


def parse_tgs_json(file):
    """!
    Reads both tgs and lottie files, returns the json structure
    """
    return open_maybe_gzipped(file, json.load)


def open_maybe_gzipped(file, on_open):
    if isinstance(file, str):
        with open(file, "r") as fileobj:
            return open_maybe_gzipped(fileobj, on_open)

    if isinstance(file, io.TextIOBase) and hasattr(file, "buffer"):
        binfile = file.buffer
    else:
        binfile = file

    try:
        binfile.seek(binfile.tell()) # Throws when not seekable
        mn = binfile.read(2)
        binfile.seek(0)
    except (io.UnsupportedOperation, OSError):
        mn = b''

    if mn == b'\x1f\x8b': # gzip magic number
        final_file = gzip.open(binfile, "rb")
    elif isinstance(file, io.TextIOBase):
        final_file = file
    else:
        final_file = io.TextIOWrapper(file)

    return on_open(final_file)


def parse_tgs(filename):
    """!
    Reads both tgs and lottie files
    """
    lottie = parse_tgs_json(filename)
    return Animation.load(lottie)



try:
    import cairosvg
    has_cairo = True
except (ImportError, OSError):
    has_cairo = False

from ..parsers import glaxnimate_helpers

import io

from .base import exporter
from .svg import export_svg


if glaxnimate_helpers.has_glaxnimate:
    @exporter("PNG", ["png"], [], {"frame"})
    def export_png(animation, fp, frame=0, dpi=96):
        data = glaxnimate_helpers.serialize(animation, "raster")

        if isinstance(fp, str):
            with open(fp, 'wb') as file:
                file.write(data)
        else:
            fp.write(data)

    def PngRenderer(animation, dpi):
        return glaxnimate_helpers.GlaxnimateRenderer(animation, "raster", dpi)

elif has_cairo:
    def _export_cairo(func, animation, fp, frame, dpi):
        intermediate = io.BytesIO()
        export_svg(animation, intermediate, frame, pretty=False)
        intermediate.seek(0)
        func(file_obj=intermediate, write_to=fp, dpi=dpi)

    @exporter("PNG", ["png"], [], {"frame"})
    def export_png(animation, fp, frame=0, dpi=96):
        _export_cairo(cairosvg.svg2png, animation, fp, frame, dpi)

    class PngRenderer:
        def __init__(self, animation, dpi):
            self.animation = animation
            self.dpi = dpi

        def __enter__(self):
            return self

        def __exit__(self, *a, **k):
            return

        def serialize(self, frame, file):
            export_png(self.animation, file, frame, self.dpi)

    @exporter("PDF", ["pdf"], [], {"frame"})
    def export_pdf(animation, fp, frame=0, dpi=96):
        _export_cairo(cairosvg.svg2pdf, animation, fp, frame, dpi)

    @exporter("PostScript", ["ps"], [], {"frame"})
    def export_ps(animation, fp, frame=0, dpi=96):
        _export_cairo(cairosvg.svg2ps, animation, fp, frame, dpi)



import os
import enum
import json
import inspect

from ..parsers.tgs import parse_tgs
from ..objects.base import ObjectVisitor
from ..objects.animation import Animation
from ..objects import layers
from ..objects import shapes
from ..objects import helpers


class Severity(enum.Enum):
    Note = enum.auto()
    Warning = enum.auto()
    Error = enum.auto()


class TgsError:
    def __init__(self, message, target, severity=Severity.Warning):
        self.message = message
        self.target = target
        self.severity = severity

    def target_id(self):
        if isinstance(self.target, str):
            return self.target
        if getattr(self.target, "name", ""):
            return self.target.name
        return self.target.__class__.__name__

    def __str__(self):
        return "%s: on %s: %s" % (
            self.severity.name,
            self.target_id(),
            self.message
        )


class TgsValidator(ObjectVisitor):
    def __init__(self, severity=Severity.Note):
        self.errors = []
        self.severity = severity

    def _check(self, expr, message, target, severity=Severity.Warning):
        if severity.value >= self.severity.value and not expr:
            self.errors.append(TgsError(message, target, severity))

    def check_file_size(self, filename):
        return self.check_size(os.path.getsize(filename))

    def check_size(self, bytes, filename="file"):
        size_k = bytes / 1024
        self._check(
            size_k <= 64,
            "Invalid size (%.1fk), should be less than 64k" % size_k,
            filename,
            Severity.Error
        )

    def check_file(self, filename):
        self.check_file_size(filename)
        try:
            self(parse_tgs(filename))
        except json.decoder.JSONDecodeError as e:
            self._check(
                False,
                "Invalid JSON: %s" % e,
                filename,
                Severity.Error
            )

    def visit(self, object):
        for cls in inspect.getmro(object.__class__):
            callback = "_visit_%s" % cls.__name__.lower()
            if hasattr(self, callback):
                getattr(self, callback)(object)

    def _visit_animation(self, o: Animation):
        self._check(
            o.frame_rate in {30, 60},
            "Invalid framerate %s, should be 30 or 60" % o.frame_rate,
            o,
            Severity.Error
        )
        self._check(
            o.width == 512,
            "Invalid width %s, should be 512" % o.width,
            o,
            Severity.Error
        )
        self._check(
            o.height == 512,
            "Invalid height %s, should be 512" % o.height,
            o,
            Severity.Error
        )
        self._check(
            (o.out_point-o.in_point) <= 180,
            "Too many frames (%s), should be less than 180" % (o.out_point-o.in_point),
            o,
            Severity.Error
        )

    def _visit_layer(self, o: layers.Layer):
        self._check(
            not o.has_masks and not o.masks,
            "Masks are not officially supported",
            o,
            Severity.Note
        )
        self._check(
            not o.effects,
            "Effects are not supported",
            o,
            Severity.Warning
        )
        self._check(
            not o.threedimensional,
            "3D layers are not supported",
            o,
            Severity.Warning
        )
        self._check(
            not isinstance(o, layers.TextLayer),
            "Text layers are not supported",
            o,
            Severity.Warning
        )
        self._check(
            not isinstance(o, layers.ImageLayer),
            "Image layers are not supported",
            o,
            Severity.Warning
        )
        self._check(
            not o.auto_orient,
            "Auto-orient layers are not supported",
            o,
            Severity.Warning
        )
        self._check(
            o.matte_mode in {None, layers.MatteMode.Normal},
            "Mattes are not supported",
            o,
            Severity.Warning
        )

    def _visit_precomplayer(self, o: layers.PreCompLayer):
        self._check(
            o.time_remapping is None,
            "Time remapping is not supported",
            o,
            Severity.Warning
        )

    def _visit_merge(self, o: shapes.Merge):
        self._check(
            False,
            "Merge paths are not supported",
            o,
            Severity.Warning
        )

    def _visit_transform(self, o: helpers.Transform):
        self._check(
            o.skew is None or (
                not o.skew.animated and o.skew.value == 0
            ),
            "Skew transforms are not supported",
            o,
            Severity.Warning
        )

    def _visit_gradientstroke(self, o: shapes.GradientStroke):
        self._check(
            False,
            "Gradient strokes are not officially supported",
            o,
            Severity.Note
        )

    def _visit_star(self, o: shapes.Star):
        self._check(
            False,
            "Star Shapes are not officially supported",
            o,
            Severity.Note
        )

    def _visit_repeater(self, o: shapes.Repeater):
        self._check(
            False,
            "Repeaters are not officially supported",
            o,
            Severity.Note
        )



from .. import objects


class TensorSerializer:
    def process(self, animation: objects.Animation, flatten: bool = True):
        data = []
        for layer in animation.layers:
            self.process_layer(layer, data)

        if flatten:
            return self.flatten(data)

        return data

    def process_layer(self, layer, data):
        if not isinstance(layer, objects.ShapeLayer):
            return

        self.process_shape_group(layer, data)

    def process_shape(self, shape: objects.Path, data: dict):
        bez = shape.shape.get_value()
        if bez and len(bez.vertices) > 1:
            data["curves"].append(bez)

    def process_styler(self, styler):
        return {
            "opacity": (styler.opacity.get_value() or 100) / 100,
            "color":
                styler.gradient.get_stops()[0][0]
                if isinstance(styler, objects.Gradient)
                else styler.color.get_value()
        }

    def process_shape_group(self, group, data):
        shape_data = {
            "curves": [],
            "fill": None,
            "stroke": None
        }

        for shape in group.shapes:
            if shape.hidden:
                continue

            if isinstance(shape, objects.Group):
                child_data = self.process_shape_group(shape, data)
                shape_data["curves"] += child_data["curves"]
            elif isinstance(shape, objects.Path):
                self.process_shape(shape, shape_data)
            elif isinstance(shape, objects.Shape):
                self.process_shape(shape.to_bezier(), shape_data)
            elif isinstance(shape, (objects.Fill, objects.GradientFill)):
                shape_data["fill"] = self.process_styler(shape)
            elif isinstance(shape, objects.BaseStroke):
                shape_data["stroke"] = self.process_styler(shape)
                shape_data["stroke"]["width"] = shape.width.get_value()

        if shape_data["curves"] and (shape_data["fill"] or shape_data["stroke"]):
            data.append(shape_data)

        return shape_data

    def flatten_color(self, data):
        if not data:
            return None
        return list(data["color"])[0:3] + [data["opacity"]]

    def flatten(self, data):
        flattened = []
        for item in data:
            base = {
                "fill": self.flatten_color(item["fill"]),
                "stroke": self.flatten_color(item["stroke"]),
                "stroke_width": item["stroke"]["width"] if item["stroke"] else 0,
            }
            base["color"] = base["fill"] or base["stroke"]

            for curve in item["curves"]:
                points = []
                for i in range(len(curve.vertices)):
                    v = curve.vertices[i]
                    points.append(list(v))
                    points.append(list(curve.out_tangents[i] + v))
                    next_i = (i+1) % len(curve.vertices)
                    points.append(list(curve.in_tangents[next_i] + curve.vertices[next_i]))

                if curve.closed:
                    points.push(list(curve.vertices[0]))
                else:
                    points.pop()
                    points.pop()

                flattened.append({
                    **base,
                    "points": points,
                    "closed": int(curve.closed)
                })

        return flattened



#!/usr/bin/env python3

import sys
import os
import argparse
sys.path.insert(0, os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from lottie.exporters import exporters
from lottie.importers import importers
from lottie.utils.stripper import float_strip, heavy_strip
from lottie import __version__


def print_loader(loader, type):
    print("* Available %s:" % type)
    for porter in loader:
        print("  * %s" % porter.name)
    if loader.failed_modules:
        print("* Failed %s:" % type)
        for name, missing in loader.failed_modules.items():
            print("  * %s (missing %s)" % (name, missing))


print("* Python version: %s" % sys.version.replace("\n", " "))
print("* Python Lottie version: %s" % __version__)
print_loader(importers, "Importers")
print_loader(exporters, "Exporters")

try:
    import glaxnimate
    print("* Glaxnimate rendering available")
except ImportError:
    pass



from xml.dom import minidom
from distutils.util import strtobool

from lottie.nvector import NVector
from lottie.parsers.sif.sif.frame_time import FrameTime


class _tag:
    def __init__(self, type):
        self.type = type

    def __call__(self, v):
        return self.type(v)


bool_str = _tag(bool)


def str_to_bool(strval):
    return bool(strtobool(strval))


def value_from_xml_string(xml_str, type, registry):
    if type in (bool_str, bool):
        return str_to_bool(xml_str)
    elif type is NVector:
        return NVector(*map(float, xml_str.split()))
    if type is FrameTime:
        return FrameTime.parse_string(xml_str, registry)
    return type(xml_str)


def value_to_xml_string(value, type):
    if type is bool:
        return "1" if value else "0"
    if type is bool_str:
        return "true" if value else "false"
    elif type is NVector:
        return " ".join(map(str, value))
    return str(value)


def value_isinstance(value, type):
    if isinstance(type, _tag):
        type = type.type
    return isinstance(value, type)


def xml_text(node):
    return "".join(
        x.nodeValue
        for x in node.childNodes
        if x.nodeType in {minidom.Node.TEXT_NODE, minidom.Node.CDATA_SECTION_NODE}
    )


def xml_make_text(dom: minidom.Document, tag_name, text):
    e = dom.createElement(tag_name)
    e.appendChild(dom.createTextNode(text))
    return e


def xml_element_matches(ch: minidom.Node, tagname=None):
    if ch.nodeType != minidom.Node.ELEMENT_NODE:
        return False

    if tagname is not None and ch.tagName != tagname:
        return False

    return True


def xml_child_elements(xml: minidom.Node, tagname=None):
    for ch in xml.childNodes:
        if xml_element_matches(ch, tagname):
            yield ch


def xml_first_element_child(xml: minidom.Node, tagname=None, allow_none=False):
    for ch in xml_child_elements(xml, tagname):
        return ch

    if allow_none:
        return None
    raise ValueError("No %s in %s" % (tagname or "child element", getattr(xml, "tagName", "node")))



from xml.dom import minidom
import enum
from uuid import uuid4

from lottie.nvector import NVector
from lottie.parsers.sif.xml.utils import xml_text, str_to_bool
from lottie.parsers.sif.xml.utils import xml_child_elements, value_from_xml_string, xml_make_text, value_to_xml_string
from lottie.parsers.sif.sif.frame_time import FrameTime


class ObjectRegistry:
    def __init__(self):
        self.registry = {}

    def register_as(self, object, key):
        self.registry[key] = object

    def register(self, object):
        guid = getattr(object, "guid", None)
        if guid is None:
            guid = self.guid()
            object.guid = guid
        self.registry[guid] = object

    @classmethod
    def guid(cls):
        return str(uuid4()).replace("-", "").upper()

    def get_object(self, guid):
        return self.registry[guid]


def noop(x):
    return x


class TypeDescriptor:
    _type_tag_names = {
        "bone_object": "bone"
    }

    def __init__(self, typename, default=None, type_wrapper=noop):
        self.typename = typename
        self.type_wrapper = type_wrapper
        self.default_value = default

    def value_to_xml_element(self, value, dom: minidom.Document):
        element = dom.createElement(self.tag_name)

        if self.typename == "vector":
            element.appendChild(xml_make_text(dom, "x", str(value.x)))
            element.appendChild(xml_make_text(dom, "y", str(value.y)))
            if hasattr(value, "guid"):
                element.setAttribute("guid", value.guid)
        elif self.typename == "color":
            element.appendChild(xml_make_text(dom, "r", str(value[0])))
            element.appendChild(xml_make_text(dom, "g", str(value[1])))
            element.appendChild(xml_make_text(dom, "b", str(value[2])))
            element.appendChild(xml_make_text(dom, "a", str(value[3])))
            if hasattr(value, "guid"):
                element.setAttribute("guid", value.guid)
        elif self.typename == "gradient":
            for point in value:
                element.appendChild(point.to_dom(dom))
        elif self.typename == "bool":
            element.setAttribute("value", "true" if value else "false")
        elif self.typename == "bone_object":
            element.setAttribute("guid", value.guid)
            element.setAttribute("type", self.typename)
        elif self.typename == "string":
            element.appendChild(dom.createTextNode(value))
        else:
            if isinstance(value, enum.Enum):
                value = value.value
            element.setAttribute("value", str(value))

        return element

    @property
    def tag_name(self):
        return self._type_tag_names.get(self.typename, self.typename)

    def value_from_xml_element(self, xml: minidom.Element, registry: ObjectRegistry):
        if xml.tagName != self.tag_name:
            raise ValueError("Wrong value type (%s instead of %s)" % (xml.tagName, self.tag_name))

        guid = xml.getAttribute("guid")
        if guid and guid in registry.registry:
            value = registry.registry[guid]
        elif self.typename == "vector":
            value = NVector(
                float(xml_text(xml.getElementsByTagName("x")[0])),
                float(xml_text(xml.getElementsByTagName("y")[0]))
            )
            if xml.getAttribute("guid"):
                value.guid = xml.getAttribute("guid")
                registry.register(value)
        elif self.typename == "color":
            value = NVector(
                float(xml_text(xml.getElementsByTagName("r")[0])),
                float(xml_text(xml.getElementsByTagName("g")[0])),
                float(xml_text(xml.getElementsByTagName("b")[0])),
                float(xml_text(xml.getElementsByTagName("a")[0]))
            )
        elif self.typename == "gradient":
            value = [
                GradientPoint.from_dom(sub, registry)
                for sub in xml_child_elements(xml, GradientPoint.type.typename)
            ]
        elif self.typename == "real" or self.typename == "angle":
            value = float(xml.getAttribute("value"))
        elif self.typename == "integer":
            value = int(xml.getAttribute("value"))
        elif self.typename == "time":
            value = FrameTime.parse_string(xml.getAttribute("value"), registry)
        elif self.typename == "bool":
            value = str_to_bool(xml.getAttribute("value"))
        elif self.typename == "string":
            return xml_text(xml)
        elif self.typename == "bone_object":
            # Already done above but this forces the guid to be present
            return registry.get_object(xml.getAttribute("guid"))
        else:
            raise ValueError("Unsupported type %s" % self.typename)

        return self.type_wrapper(value)


class GradientPoint:
    type = TypeDescriptor("color")

    def __init__(self, pos: float, color: NVector):
        self.pos = pos
        self.color = color

    def to_dom(self, dom: minidom.Document):
        element = self.type.value_to_xml_element(self.color, dom)
        element.setAttribute("pos", value_to_xml_string(self.pos, float))
        return element

    @classmethod
    def from_dom(cls, xml: minidom.Element, registry: ObjectRegistry):
        return GradientPoint(
            value_from_xml_string(xml.getAttribute("pos"), float, registry),
            cls.type.value_from_xml_element(xml, registry)
        )

    def __repr__(self):
        return "<GradientPoint %s %s>" % (self.pos, self.color)


class SifNodeMeta(type):
    def __new__(cls, name, bases, attr):
        props = []
        for base in bases:
            if type(base) == cls:
                props += base._nodes
        attr["_nodes"] = props + attr.get("_nodes", [])
        if "_tag" not in attr:
            attr["_tag"] = name.lower()
        attr["_nodemap"] = {
            node.att_name: node
            for node in attr["_nodes"]
        }
        return super().__new__(cls, name, bases, attr)



from . import base, core, sif, svg
from .base import importers

__all__ = [
    "base", "core", "sif", "svg",
    "importers",
]

try:
    from . import raster
    __all__ += ["raster"]
except ImportError:
    pass



from PyQt5 import QtWidgets
from PyQt5.QtGui import *
from PyQt5.QtCore import Qt

from .. import objects


def lottie_theme_icon(lottie_object):
    if isinstance(lottie_object, objects.Animation):
        return "tool-animator"

    if isinstance(lottie_object, objects.Precomp):
        return "folder"

    if isinstance(lottie_object, objects.ShapeLayer):
        return "shapes"
    if isinstance(lottie_object, objects.TextLayer):
        return "draw-text"
    if isinstance(lottie_object, objects.PreCompLayer):
        return "emblem-symbolic-link"

    if isinstance(lottie_object, (objects.Rect, objects.SolidColorLayer)):
        return "draw-rectangle"
    if isinstance(lottie_object, objects.Ellipse):
        return "draw-ellipse"
    if isinstance(lottie_object, objects.Star):
        return "draw-star"
    if isinstance(lottie_object, objects.Path):
        return "draw-bezier-curves"
    if isinstance(lottie_object, (objects.Fill, objects.GradientFill)):
        return "format-fill-color"
    if isinstance(lottie_object, (objects.Stroke, objects.GradientStroke)):
        return "format-stroke-color"
    if isinstance(lottie_object, objects.Transform):
        return "transform-scale"
    if isinstance(lottie_object, objects.Group):
        return "object-group"

    return None


def lottie_to_tree(tree_parent, lottie_object):
    prop_to_tree(tree_parent, None, lottie_object)


def lottie_object_to_tree(item, lottie_object, propname, textcol):
    if propname == "color" and isinstance(lottie_object, objects.MultiDimensional) and not lottie_object.animated:
        item.setBackground(textcol, QBrush(QColor.fromRgbF(*lottie_object.value)))
    else:
        text = str(lottie_object)
        if text != propname or textcol == 0:
            item.setText(textcol, text)

    icon = lottie_theme_icon(lottie_object)
    if icon:
        item.setIcon(0, QIcon.fromTheme(icon))

    for prop in lottie_object._props:
        propitem = prop_to_tree(item, prop.name, prop.get(lottie_object))
        propitem.setToolTip(0, prop.lottie)


def prop_to_tree(tree_parent, propname, propval):
    item = QtWidgets.QTreeWidgetItem(tree_parent)

    first_column = 0
    if propname:
        item.setText(0, propname)
        first_column = 1

    if isinstance(propval, objects.LottieObject):
        lottie_object_to_tree(item, propval, propname, first_column)
    elif isinstance(propval, list):
        for subval in propval:
            prop_to_tree(item, type(subval).__name__, subval)
    elif propval is None:
        item.setText(first_column, "")
    elif isinstance(propval, bool):
        item.setCheckState(first_column, Qt.CheckState.Checked if propval else Qt.CheckState.Unchecked)
        item.setFlags(Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled)
    else:
        item.setText(first_column, str(propval))

    return item



from xml.dom import minidom
import enum

from lottie.nvector import NVector
from lottie.parsers.sif.ast_impl.base import SifAstNode, TypeDescriptor, ObjectRegistry
from lottie.parsers.sif.xml.animatable import XmlAnimatable
from lottie.parsers.sif.xml.wrappers import XmlBoneReference, XmlSifElement, XmlList
from lottie.parsers.sif.sif.nodes import Segment, WeightedVector, Bline
from lottie.parsers.sif.sif.enums import Smooth
from lottie.parsers.sif.sif.core import SifNodeMeta, FrameTime


class SifAstComplex(SifAstNode, metaclass=SifNodeMeta):
    _nodes = []

    def __init__(self, **kw):
        for node in self._nodes:
            node.initialize_object(kw, self)

    @classmethod
    def from_dom(cls, xml: minidom.Element, param: TypeDescriptor, registry: ObjectRegistry):
        outcls = cls.get_class_from_dom(xml, param, registry)
        instance = outcls()
        for node in outcls._nodes:
            node.from_xml(instance, xml, registry)
        return instance

    @classmethod
    def get_class_from_dom(cls, xml: minidom.Element, param: TypeDescriptor, registry: ObjectRegistry):
        return cls

    def to_dom(self, dom: minidom.Document, param: TypeDescriptor):
        element = self._prepare_to_dom(dom, param)
        for node in self._nodes:
            node.to_xml(self, element, dom, param)
        return element


class SifAstBoneLink(SifAstComplex):
    _tag = "bone_link"

    _nodes = [
        XmlBoneReference("bone"),
        XmlAnimatable("base_value", "vector", NVector(0, 0)),
        XmlAnimatable("translate", "bool", True),
        XmlAnimatable("rotate", "bool", True),
        XmlAnimatable("skew", "bool", True),
        XmlAnimatable("scale_x", "bool", True),
        XmlAnimatable("scale_y", "bool", True),
    ]


class SifAstBoneInfluence(SifAstComplex):
    _tag = "boneinfluence"

    _nodes = [
        # TODO bone_weight_list
        XmlAnimatable("link", "vector", NVector(0, 0)),
    ]


class SifSegCalcTangent(SifAstComplex):
    _tag = "segcalctangent"

    _nodes = [
        XmlSifElement("segment", Segment),
        XmlAnimatable("amount", "real", .5),
    ]


class SifSegCalcVertex(SifAstComplex):
    _tag = "segcalcvertex"

    _nodes = [
        XmlSifElement("segment", Segment),
        XmlAnimatable("amount", "real", .5),
    ]


class WeightedAverage(SifAstComplex):
    _tag = "weighted_average"

    _nodes = [
        XmlList(WeightedVector, "vectors", "entry"),
    ]

    def _prepare_to_dom(self, dom: minidom.Document, param: TypeDescriptor):
        element = dom.createElement(self._tag)
        element.setAttribute("type", "weighted_vector")
        return element


class SifAdd(SifAstComplex):
    _tag = "add"

    _nodes = [
        XmlAnimatable("lhs", "_recurse"),
        XmlAnimatable("rhs", "_recurse"),
        XmlAnimatable("scalar", "real", 1.),
    ]


class SifAnimatedFile(SifAstComplex):
    _tag = "animated_file"

    _nodes = [
        XmlAnimatable("filename", "string"),
    ]


class Accuracy(enum.Enum):
    Rough = 0
    Normal = 1
    Fine = 2
    Extreme = 3


class DerivativeOrder:
    FirstDerivative = 0
    SecondDerivative = 1


class SifDerivative(SifAstComplex):
    _tag = "derivative"

    _nodes = [
        XmlAnimatable("link", "_recurse"),
        XmlAnimatable("interval", "real", 0.01),
        XmlAnimatable("accuracy", "integer", Accuracy.Normal, Accuracy),
        XmlAnimatable("order", "integer", DerivativeOrder.FirstDerivative, DerivativeOrder),
    ]


class SifDynamic(SifAstComplex):
    _tag = "dynamic"

    _nodes = [
        XmlAnimatable("tip_static", "vector", NVector(0, 0)),
        XmlAnimatable("origin", "vector", NVector(0, 0)),
        XmlAnimatable("force", "vector", NVector(0, 0)),
        XmlAnimatable("torque", "real", 0.),
        XmlAnimatable("damping", "real", 0.4),
        XmlAnimatable("friction", "real", 0.4),
        XmlAnimatable("spring", "real", 30.),
        XmlAnimatable("torsion", "real", 30.),
        XmlAnimatable("mass", "real", 0.3),
        XmlAnimatable("inertia", "real", 0.3),
        XmlAnimatable("spring_rigid", "bool", False),
        XmlAnimatable("torsion_rigid", "bool", False),
        XmlAnimatable("origin_drags_tip", "bool", True),
    ]


class SifGreyed(SifAstComplex):
    _tag = "greyed"

    _nodes = [
        XmlAnimatable("link", "_recurse"),
    ]


class SifLinear(SifAstComplex):
    _tag = "linear"

    _nodes = [
        XmlAnimatable("slope", "vector", NVector(0, 0)),
        XmlAnimatable("offset", "vector", NVector(0, 0)),
    ]


class SifRadialComposite(SifAstComplex):
    _tag = "radial_composite"

    _nodes = [
        XmlAnimatable("radius", "real", 0.),
        XmlAnimatable("theta", "angle", 0.),
    ]


class SifComposite(SifAstComplex):
    _tag = "composite"

    @classmethod
    def get_class_from_dom(cls, xml: minidom.Element, param: TypeDescriptor, registry: ObjectRegistry):
        type = xml.getAttribute("type")
        if type == "vector":
            return SifVectorComposite
        return None

    def _prepare_to_dom(self, dom: minidom.Document, param: TypeDescriptor):
        element = dom.createElement("composite")
        element.setAttribute("type", param.typename)
        return element


class SifVectorComposite(SifComposite):
    _type = "vector"

    _nodes = [
        XmlAnimatable("x", "real", 0.),
        XmlAnimatable("y", "real", 0.),
    ]


class SifRandom(SifAstComplex):
    _tag = "random"

    _nodes = [
        XmlAnimatable("link", "_recurse"),
        XmlAnimatable("radius", "real", 0.),
        XmlAnimatable("seed", "integer", 0),
        XmlAnimatable("speed", "real", 1.),
        XmlAnimatable("smooth", "integer", Smooth.Cubic, Smooth),
        XmlAnimatable("loop", "real", 0.),
    ]


class SifReference(SifAstComplex):
    _tag = "link"

    _nodes = [
        XmlAnimatable("reference", "_recurse"),
    ]


class SifScale(SifAstComplex):
    _tag = "scale"

    _nodes = [
        XmlAnimatable("link", "_recurse"),
        XmlAnimatable("scalar", "real", 1.),
    ]


class SifStep(SifAstComplex):
    _tag = "step"

    _nodes = [
        XmlAnimatable("link", "_recurse"),
        XmlAnimatable("duration", "time", FrameTime(1, FrameTime.Unit.Seconds)),
        XmlAnimatable("start_time", "time", FrameTime(0, FrameTime.Unit.Seconds)),
        XmlAnimatable("intersection", "real", 0.5),
    ]


class SifSubtract(SifAstComplex):
    _tag = "subtract"

    _nodes = [
        XmlAnimatable("lhs", "_recurse"),
        XmlAnimatable("rhs", "_recurse"),
        XmlAnimatable("scalar", "real", 1.),
    ]


class SifSwitch(SifAstComplex):
    _tag = "switch"

    _nodes = [
        XmlAnimatable("link_off", "_recurse"),
        XmlAnimatable("link_on", "_recurse"),
        XmlAnimatable("switch", "bool", False),
    ]


class SifTimedSwap(SifAstComplex):
    _tag = "timed_swap"

    _nodes = [
        XmlAnimatable("before", "_recurse"),
        XmlAnimatable("after", "_recurse"),
        XmlAnimatable("time", "time", FrameTime(0, FrameTime.Unit.Seconds)),
        XmlAnimatable("length", "time", FrameTime(0, FrameTime.Unit.Seconds)),
    ]


class SifTimeLoop(SifAstComplex):
    _tag = "timeloop"

    _nodes = [
        XmlAnimatable("link", "_recurse"),
        XmlAnimatable("link_time", "time", FrameTime(0, FrameTime.Unit.Seconds)),
        XmlAnimatable("local_time", "time", FrameTime(0, FrameTime.Unit.Seconds)),
        XmlAnimatable("duration", "time", FrameTime(0, FrameTime.Unit.Seconds)),
    ]


class SifPower(SifAstComplex):
    _tag = "power"

    _nodes = [
        XmlAnimatable("base", "real", 1.),
        XmlAnimatable("power", "real", 1.),
        XmlAnimatable("epsilon", "real", 0.000001),
        XmlAnimatable("infinite", "real", 999999.),
    ]


class SifBlineCalcTangent(SifAstComplex):
    _tag = "blinecalctangent"

    _nodes = [
        XmlSifElement("bline", Bline),
        XmlAnimatable("loop", "bool", False),
        XmlAnimatable("amount", "real", 0.5),
        XmlAnimatable("offset", "angle", 0.),
        XmlAnimatable("scale", "real", 1.),
        XmlAnimatable("fixed_length", "bool", False),
        XmlAnimatable("homogeneous", "bool", False),
    ]


class SifBlineCalcVertex(SifAstComplex):
    _tag = "blinecalcvertex"

    _nodes = [
        XmlSifElement("bline", Bline),
        XmlAnimatable("loop", "bool", False),
        XmlAnimatable("amount", "real", 0.5),
        XmlAnimatable("homogeneous", "bool", False),
    ]



from ..nvector import NVector


# FABRIK
class Chain:
    def __init__(self, tail, fixed_tail=True, tolerance=0.5, max_iter=8):
        self.joints = [tail.clone()]
        self.fixed_tail = fixed_tail
        self.lengths = []
        self.total_length = 0
        self.tolerance = tolerance
        self.max_iter = max_iter

    def add_joint(self, point):
        length = (point - self.joints[-1]).length
        self.lengths.append(length)
        self.total_length += length
        self.joints.append(point.clone())

    def add_joints(self, head, n):
        delta = head - self.joints[-1]
        self.total_length += delta.length
        segment = delta / n
        seglen = segment.length
        for i in range(n):
            self.lengths.append(seglen)
            self.joints.append(self.joints[-1] + segment)

    def backward(self, target):
        """!
        target -> -> start
        """
        self.joints[-1] = target
        for i in range(len(self.joints)-2, -1, -1):
            r = self.joints[i+1] - self.joints[i]
            l = self.lengths[i] / r.length
            self.joints[i] = self.joints[i+1].lerp(self.joints[i], l)

    def forward(self, target):
        """!
        start -> -> tail
        """
        self.joints[0] = target
        for i in range(0, len(self.joints)-1):
            r = self.joints[i+1] - self.joints[i]
            l = self.lengths[i] / r.length
            self.joints[i+1] = self.joints[i].lerp(self.joints[i+1], l)

    def reach(self, target):
        if not self.fixed_tail:
            self.backward(target)
            return

        distance = (target - self.joints[0]).length
        if distance >= self.total_length:
            for i in range(len(self.joints)-1):
                r = target - self.joints[i]
                l = self.lengths[i] / r.length
                self.joints[i+1] = self.joints[i].lerp(target, l)
            return

        base = self.joints[0]

        distance = (target - self.joints[-1]).length
        n_it = 0
        while distance > self.tolerance and n_it < self.max_iter:
            self.backward(target)
            self.forward(base)
            distance = (target - self.joints[-1]).length
            n_it += 1


class Octopus:
    def __init__(self, master):
        self.chains = {"master": master}
        self.master = master

    @property
    def base(self):
        return self.master.joints[-1]

    def add_chain(self, name):
        ch = Chain(self.base)
        self.chains[name] = ch
        return ch

    def reach(self, target_map):
        centroid = NVector(0, 0)
        for chain, target in target_map.items():
            self.chains[chain].backward(target)
            centroid += self.chains[chain].joints[0]
        centroid /= len(target_map)

        self.master.reach(centroid)

        for chain in target_map.keys():
            self.chains[chain].forward(self.base)



import sys

from ..objects.base import LottieObject, LottieBase
from ..objects.properties import MultiDimensional, Value, ShapeProperty
from ..objects.layers import Layer


def _prettyprint_scalar(lottie_object, out=sys.stdout):
    if isinstance(lottie_object, float) and lottie_object == round(lottie_object):
        lottie_object = int(lottie_object)
    return str(lottie_object)


def prettyprint(lottie_object, out=sys.stdout, indent="   ", _i=""):
    if isinstance(lottie_object, LottieObject):
        out.write(lottie_object.__class__.__name__)
        out.write('\n')
        _i += indent
        maxk = max(map(lambda x: len(x.name), lottie_object._props))
        for k in lottie_object._props:
            out.write(_i)
            out.write(k.name.ljust(maxk))
            out.write(' : ')
            prettyprint(k.get(lottie_object), out, indent, _i)
    elif isinstance(lottie_object, (list, tuple)):
        if not lottie_object or (not isinstance(lottie_object[0], LottieBase) and len(lottie_object) < 16):
            out.write("[")
            out.write(", ".join(map(_prettyprint_scalar, lottie_object)))
            out.write("]\n")
        else:
            out.write("[\n")
            for k in lottie_object:
                out.write(_i + indent)
                prettyprint(k, out, indent, _i + indent)
            out.write(_i)
            out.write(']\n')
    else:
        out.write(_prettyprint_scalar(lottie_object, out))
        out.write('\n')


def _prettyprint_summary_printable(obj):
    if isinstance(obj, LottieObject):
        return not isinstance(obj, (MultiDimensional, Value, ShapeProperty))
    return obj and isinstance(obj, (list, tuple)) and isinstance(obj[0], LottieObject)


def prettyprint_summary(lottie_object, out=sys.stdout, indent="   ", _i=""):
    if isinstance(lottie_object, LottieObject):
        out.write(lottie_object.__class__.__name__)
        name = getattr(lottie_object, "name", None)
        if name:
            out.write(" %r" % name)
        if isinstance(lottie_object, Layer):
            out.write(" %s -> %s" % (lottie_object.index, lottie_object.parent_index))
        out.write('\n')
        _i += indent
        for k in lottie_object._props:
            val = k.get(lottie_object)
            if _prettyprint_summary_printable(val):
                out.write(_i)
                out.write(k.name)
                out.write(' : ')
                prettyprint_summary(val, out, indent, _i)
    elif _prettyprint_summary_printable(lottie_object):
            out.write("[\n")
            for k in lottie_object:
                out.write(_i + indent)
                prettyprint_summary(k, out, indent, _i + indent)
            out.write(_i)
            out.write(']\n')



import re
import sys
import enum
import math
import typing

from ..parsers.svg.svgdata import color_table
from ..parsers.svg.importer import parse_color, parse_svg_file
from ..objects.animation import Animation
from ..objects import layers
from ..objects import shapes
from ..nvector import NVector
from .font import FontStyle


color_words = {
    "alice": {"blue": {"_": 1}},
    "antique": {"white": {"_": 1}},
    "aqua": {
        "_": 1,
        "marine": {"_": 1}
    },
    "azure": {"_": 1},
    "beige": {"_": 1},
    "bisque": {"_": 1},
    "black": {"_": 1},
    "blanched": {"almond": {"_": 1}},
    "blue": {
        "_": 1,
        "navy": "navy",
        "violet": {"_": 1},
    },
    "brown": {"_": 1},
    "burly": {"wood": {"_": 1}},
    "cadet": {"blue": {"_": 1}},
    "chartreuse": {"_": 1},
    "chocolate": {"_": 1},
    "coral": {"_": 1},
    "cornflower": {"blue": {"_": 1}},
    "corn": {
        "silk": {"_": 1},
        "flower": {"blue": {"_": 1}},
    },
    "crimson": {"_": 1},
    "cyan": {"_": 1},
    "dark": {
        "blue": {"_": 1},
        "cyan": {"_": 1},
        "golden": {"rod": {"_": 1}},
        "gray": {"_": 1},
        "green": {"_": 1},
        "grey": {"_": 1},
        "khaki": {"_": 1},
        "magenta": {"_": 1},
        "olive": {"green": {"_": 1}},
        "orange": {"_": 1},
        "orchid": {"_": 1},
        "red": {"_": 1},
        "salmon": {"_": 1},
        "sea": {"green": {"_": 1}},
        "slate": {
            "blue": {"_": 1},
            "gray": {"_": 1},
            "grey": {"_": 1}
        },
        "turquoise": {"_": 1},
        "violet": {"_": 1}
    },
    "deep": {
        "pink": {"_": 1},
        "sky": {"blue": {"_": 1}},
    },
    "dim": {
        "gray": {"_": 1},
        "grey": {"_": 1}
    },
    "dodger": {"blue": {"_": 1}},
    "fire": {"brick": {"_": 1}},
    "floral": {"white": {"_": 1}},
    "forest": {"green": {"_": 1}},
    "fuchsia": {"_": 1},
    "gainsboro": {"_": 1},
    "ghost": {"white": {"_": 1}},
    "gold": {"_": 1},
    "golden": {"rod": {"_": 1}},
    "gray": {"_": 1},
    "green": {
        "_": 1,
        "yellow": {"_": 1}
    },
    "grey": {"_": 1},
    "honeydew": {"_": 1},
    "hotpink": {"_": 1},
    "indian": {"red": {"_": 1}},
    "indigo": {"_": 1},
    "ivory": {"_": 1},
    "khaki": {"_": 1},
    "lavender": {
        "_": 1,
        "blush": {"_": 1},
    },
    "lawn": {"green": {"_": 1}},
    "lemon": {"chiffon": {"_": 1}},
    "light": {
        "blue": {"_": 1},
        "coral": {"_": 1},
        "cyan": {"_": 1},
        "golden": {
            "rod": {
                "_": 1,
                "yellow": {"_": 1},
            }
        },
        "gray": {"_": 1},
        "green": {"_": 1},
        "grey": {"_": 1},
        "pink": {"_": 1},
        "salmon": {"_": 1},
        "sea": {"green": {"_": 1}},
        "sky": {"blue": {"_": 1}},
        "slate": {
            "gray": {"_": 1},
            "grey": {"_": 1}
        },
        "steel": {"blue": {"_": 1}},
        "yellow": {"_": 1},
    },
    "lime": {
        "_": 1,
        "green": {"_": 1},
    },
    "linen": {"_": 1},
    "magenta": {"_": 1},
    "maroon": {"_": 1},
    "medium": {
        "aquamarine": {"_": 1},
        "blue": {"_": 1},
        "orchid": {"_": 1},
        "purple": {"_": 1},
        "sea": {"green": {"_": 1}},
        "slate": {"blue": {"_": 1}},
        "spring": {"green": {"_": 1}},
        "turquoise": {"_": 1},
        "violet": {"red": {"_": 1}},
    },
    "midnight": {"blue": {"_": 1}},
    "mint": {"cream": {"_": 1}},
    "misty": {"rose": {"_": 1}},
    "moccasin": {"_": 1},
    "navajo": {"white": {"_": 1}},
    "navy": {
        "_": 1,
        "blue": "navy",
    },
    "old": {"lace": {"_": 1}},
    "olive": {
        "_": 1,
        "drab": {"_": 1},
    },
    "orange": {
        "_": 1,
        "red": {"_": 1},
    },
    "orchid": {"_": 1},
    "pale": {
        "golden": {"rod": {"_": 1}},
        "green": {"_": 1},
        "turquoise": {"_": 1},
        "violet": {"red": {"_": 1}},
    },
    "papaya": {"whip": {"_": 1}},
    "peach": {"puff": {"_": 1}},
    "peru": {"_": 1},
    "pink": {"_": 1},
    "plum": {"_": 1},
    "powder": {"blue": {"_": 1}},
    "purple": {"_": 1},
    "red": {"_": 1},
    "rosy": {"brown": {"_": 1}},
    "royal": {"blue": {"_": 1}},
    "saddle": {"brown": {"_": 1}},
    "salmon": {"_": 1},
    "sandy": {"brown": {"_": 1}},
    "sea": {
        "green": {"_": 1},
        "shell": {"_": 1},
    },
    "seashell": {"_": 1},
    "sienna": {"_": 1},
    "silver": {"_": 1},
    "sky": {"blue": {"_": 1}},
    "slate": {
        "blue": {"_": 1},
        "gray": {"_": 1},
        "grey": {"_": 1},
    },
    "snow": {"_": 1},
    "spring": {"green": {"_": 1}},
    "steel": {"blue": {"_": 1}},
    "tan": {"_": 1},
    "teal": {"_": 1},
    "thistle": {"_": 1},
    "tomato": {"_": 1},
    "turquoise": {"_": 1},
    "violet": {"_": 1},
    "wheat": {"_": 1},
    "white": {
        "_": 1,
        "smoke": {"_": 1},
    },
    "yellow": {
        "_": 1,
        "green": {"_": 1},
    },
}


class TokenType(enum.Enum):
    Word = enum.auto()
    Number = enum.auto()
    Eof = enum.auto()
    String = enum.auto()
    Punctuation = enum.auto()
    Color = enum.auto()


class Token:
    def __init__(
        self,
        type: TokenType,
        line: int,
        col: int,
        start: int,
        end: int,
        value: typing.Any = None
    ):
        self.col = col
        self.line = line
        self.end = end
        self.start = start
        self.type = type
        self.value = value

    def __repr__(self):
        if self.type == TokenType.Eof:
            return "End of file"
        return repr(self.value)


class AnimatableType:
    Position = enum.auto()
    Color = enum.auto()
    Angle = enum.auto()
    Size = enum.auto()
    Number = enum.auto()
    Integer = enum.auto()


class AnimatableProperty:
    def __init__(self, name, phrase, props, type: AnimatableType):
        self.name = name.split()
        self.phrase = phrase
        self.props = props
        self.type = type

    def set_initial(self, value):
        for prop in self.props:
            prop.value = value

    def get_initial(self):
        return self.props[0].get_value(0)

    def add_keyframe(self, time, value):
        for prop in self.props:
            if not prop.animated and time > 0:
                prop.add_keyframe(0, prop.value)
            prop.add_keyframe(time, value)

    def loop(self, time):
        for prop in self.props:
            if prop.animated:
                prop.add_keyframe(time, prop.get_value(0))

    def get_value(self, time):
        return self.props[0].get_value(time)


class ShapeData:
    def __init__(self, extent):
        self.color = [0, 0, 0]
        self.color_explicit = False
        self.extent = extent
        self.size_multiplitier = 1
        self.portrait = False
        self.roundness = 0
        self.opacity = 1
        self.count = 1
        self.stroke = None
        self.stroke_on_top = True
        self.properties = {}

    def scale(self, multiplier):
        self.size_multiplitier *= multiplier
        self.extent *= multiplier

    def define_property(self, name, type, props, phrase=None):
        self.properties[name] = AnimatableProperty(name, phrase, props, type)

    def add_property(self, name, prop):
        self.properties[name].props.append(prop)


class Lexer:
    expression = re.compile(
        r'[\r\t ]*(?P<token>(?P<punc>[:,;.])|(?P<word>[a-zA-Z\']+)|(?P<color>#(?:[a-fA-F0-9]{3}){1,2})|' +
        r'(?P<number>-?[0-9]+(?P<fraction>\.[0-9]+)?)|(?P<string>"(?:[^"\\]|\\["\\nbt])+"))[\r\t ]*'
    )

    def __init__(self, text):
        self.text = text
        self.pos = 0
        self.token = None
        self.line = 1
        self.line_pos = 0
        self.tokens = []
        self.token_index = 0

    def _new_token(self, token: Token):
        self.token = token
        self.tokens.append(token)
        self.token_index = len(self.tokens)
        return token

    def next(self):
        if self.token_index < len(self.tokens):
            self.token = self.tokens[self.token_index]
            self.token_index += 1
            return self.token

        while True:
            if self.pos >= len(self.text):
                return self._new_token(Token(TokenType.Eof, self.line, self.pos - self.line_pos, self.pos, self.pos))

            if self.text[self.pos] == '\n':
                self.line += 1
                self.pos += 1
                self.line_pos = self.pos
            else:
                match = self.expression.match(self.text, self.pos)
                if match:
                    break

                self.pos += 1

        if match.group("word"):
            type = TokenType.Word
            value = match.group("word").lower()
        elif match.group("number"):
            type = TokenType.Number
            if match.group("fraction"):
                value = float(match.group("number"))
            else:
                value = int(match.group("number"))
        elif match.group("string"):
            type = TokenType.String
            value = match.group("string")[1:-1]
        elif match.group("punc"):
            value = match.group("punc")
            type = TokenType.Punctuation
        elif match.group("color"):
            value = match.group("color")
            type = TokenType.Color

        self.pos = match.end("token")
        return self._new_token(Token(type, self.line, self.pos - self.line_pos, match.start("token"), self.pos, value))

    def back(self):
        if self.token_index > 0:
            self.restore(self.token_index - 1)

    def save(self):
        return self.token_index

    def restore(self, index):
        self.token_index = index
        if index > 0:
            self.token = self.tokens[self.token_index-1]
        else:
            self.token = self.tokens[0]


class Logger:
    def warn(self, message):
        sys.stderr.write(message)
        sys.stderr.write("\n")


class DummyLogger(Logger):
    def warn(self, message):
        pass


class StorageLogger(Logger):
    def __init__(self):
        self.messages = []

    def warn(self, message):
        self.messages.append(message)


class Parser:
    sides = {
        "penta": 5,
        "hexa": 6,
        "hepta": 7,
        "octa": 8,
        "ennea": 9,
        "deca": 10,
    }

    def __init__(self, text, logger: Logger):
        self.lexer = Lexer(text)
        self.lexer.next()
        self.logger = logger
        self.allow_resize = True
        self.max_duration = None
        self.svg_shapes = []
        self.font = FontStyle("Ubuntu", 80)

    def next(self):
        return self.lexer.next()

    @property
    def token(self):
        return self.lexer.token

    def color(self):
        return self.get_color(color_words, "")

    def get_color_value(self, value, complete_word: str):
        if isinstance(value, str):
            return NVector(*color_table[value])
        elif isinstance(value, (list, tuple)):
            return NVector(*value)
        elif isinstance(value, NVector):
            return value
        else:
            return NVector(*color_table[complete_word])

    def complete_color(self, word_dict: dict, complete_word: str):
        if "_" in word_dict:
            return self.get_color_value(complete_word)
        else:
            next_item = next(iter(word_dict.item()))
            return self.complete_color(next_item[1], complete_word + next_item[0])

    def get_color(self, word_dict: dict, complete_word: str):
        if self.token.type == TokenType.Color:
            color = parse_color(self.token.value)
            self.next()
            return color

        if self.token.type != TokenType.Word:
            return None

        value = word_dict.get(self.token.value, None)
        if not value:
            return None

        if isinstance(value, dict):
            next_word = complete_word + self.token.value
            self.next()
            color = self.get_color(value, next_word)
            if color is not None:
                return color

            if "_" in value:
                return self.get_color_value(value["_"], next_word)

            self.warn("Incomplete color name")
        else:
            return self.get_color_value(value, complete_word)

    def warn(self, message):
        token = self.token
        self.logger.warn("At line %s column %s, near %r: %s" % (token.line, token.col, token, message))

    def parse(self):
        self.lottie = Animation(180, 60)
        if self.article():
            if self.check_words("animation", "composition"):
                self.next()
                self.animation()
            else:
                self.lexer.back()
        self.layers(self.lottie)
        if self.token.type != TokenType.Eof:
            self.warn("Extra tokens")
        return self.lottie

    def article(self):
        if self.check_words("a", "an", "the"):
            self.next()
            return True
        return False

    def check_words(self, *words):
        if self.token.type != TokenType.Word:
            return False

        return self.token.value in words

    def skip_words(self, *words):
        if self.check_words(*words):
            self.next()
            return True
        return False

    def require_one_of(self, *words):
        if self.check_words(*words):
            return True

        self.warn("Expected " + repr(words[0]))
        return False

    def check_word_sequence(self, words):
        token = self.token
        index = self.lexer.save()
        for word in words:
            if token.type != TokenType.Word or token.value != word:
                break
            token = self.next()
        else:
            return True

        self.lexer.restore(index)
        return False

    def possesive(self):
        return self.skip_words("its", "his", "her", "their")

    def properties(self, shape_data, callback, callback_args=[], words=["with"]):
        lexind_and = -1
        while True:
            must_have_property = self.skip_words(*words)

            self.article() or self.possesive()

            lexind = self.lexer.save()

            if not callback(*callback_args):
                self.lexer.restore(lexind)

                if shape_data and not self.shape_common_property(shape_data):
                    if must_have_property:
                        self.warn("Unknown property")
                        break
                if lexind_and != -1:
                    self.lexer.restore(lexind_and)
                break

            lexind_and = self.lexer.save()
            if not self.skip_and():
                break

    def simple_properties_callback(self, object, properties):
        if self.check_words(*properties.keys()):
            prop = self.token.value
            self.next()
            if self.check_words("of"):
                self.next()

            value = properties[prop](getattr(object, prop))
            setattr(object, prop, value)
            return True
        return False

    def animation(self):
        while True:
            if self.check_words("lasts", "lasting"):
                self.next()
                if self.check_words("for"):
                    self.next()
                self.lottie.out_point = self.time(self.lottie.out_point)
            elif self.check_words("stops", "stopping", "loops", "looping"):
                if self.check_words("for", "after"):
                    self.next()
                self.lottie.out_point = self.time(self.lottie.out_point)
                if self.max_duration and self.lottie.out_point > self.max_duration:
                    self.lottie.out_point = self.max_duration
            elif self.check_words("with", "has"):
                props = {
                    "width": self.integer,
                    "height": self.integer,
                    "name": self.string
                }
                if not self.allow_resize:
                    props.pop("width")
                    props.pop("height")
                self.properties(None, self.simple_properties_callback, [self.lottie, props], ["with", "has"])
            elif self.skip_and():
                pass
            else:
                return

    def time(self, default):
        if self.token.type != TokenType.Number:
            self.warn("Expected time")
            return default

        amount = self.token.value

        self.next()
        if self.check_words("seconds", "second"):
            amount *= self.lottie.frame_rate
        elif self.check_words("milliseconds", "millisecond"):
            amount *= self.lottie.frame_rate / 1000
        elif not self.check_words("frames", "frame"):
            self.warn("Missing time unit")
            return amount

        self.next()
        return amount

    def integer(self, default, warn=True):
        if self.token.type != TokenType.Number or not isinstance(self.token.value, int):
            if warn:
                self.warn("Expected integer")
            return default

        val = self.token.value
        self.next()
        return val

    def number(self, default):
        if self.token.type != TokenType.Number:
            self.warn("Expected number")
            return default

        val = self.token.value
        self.next()

        return val

    def string(self, default):
        if self.token.type != TokenType.String:
            self.warn("Expected string")
            return default

        val = self.token.value
        self.next()
        return val

    def layers(self, composition):
        while True:
            if self.token.type == TokenType.Punctuation and self.token.value in ";.":
                self.next()
            self.skip_and()
            self.skip_words("then") or self.skip_words("finally")

            if self.check_words("there's"):
                self.next()
                self.layer(composition)
            elif self.check_words("there"):
                self.next()
                if self.check_words("is", "are"):
                    self.next()
                    self.layer(composition)
            elif self.article():
                self.lexer.back()
                self.layer(composition)
            elif self.token.type == TokenType.Number and isinstance(self.token.value, int):
                self.layer(composition)
            else:
                break

    def count(self, default=1):
        if self.article():
            return 1
        return self.integer(default)

    def layer(self, composition):
        if self.token.type != TokenType.Word:
            self.warn("Expected shape")
            return

        layer = layers.ShapeLayer()
        layer.in_point = self.lottie.in_point
        layer.out_point = self.lottie.out_point
        composition.insert_layer(0, layer)

        self.shape_list(layer)

        if self.token.type == TokenType.Punctuation and self.token.value in ",;.":
            self.next()

    def skip_and(self):
        if self.token.type == TokenType.Punctuation and self.token.value == ",":
            self.next()
            self.skip_words("and")
            return True
        return self.skip_words("and")

    def shape_list(self, parent):
        extent = min(self.lottie.width, self.lottie.height) * 0.4
        shape = ShapeData(extent)
        shape.count = self.count()

        while True:
            ok = False

            color = self.color()
            if color:
                ok = True
                shape.color = color
                shape.color_explicit = True

            if self.check_words("transparent", "invisible"):
                self.next()
                shape.color = None
                ok = True

            size_mult = self.size_multiplitier()
            if size_mult:
                ok = True
                shape.scale(size_mult)

            if self.check_words("portrait"):
                self.next()
                shape.portrait = True
                ok = True

            if self.check_words("landscape"):
                self.next()
                shape.portrait = False
                ok = True

            lexind = self.lexer.save()
            qualifier = self.size_qualifier()
            if self.check_words("rounded"):
                shape.roundness = qualifier
                self.next()
                ok = True
            elif self.check_words("transparent"):
                shape.opacity = (1 / qualifier)
                self.next()
                ok = True
            elif lexind < self.lexer.token_index:
                self.lexer.restore(lexind)

            if self.check_words("star", "polygon", "ellipse", "rectangle", "circle", "square", "text"):
                shape_type = self.token.value
                function = getattr(self, "shape_" + shape_type)
                self.next()
                shape_object = function(shape)
                self.add_shape(parent, shape_object, shape)
                return

            if self.token.type == TokenType.Word:
                for name, sides in self.sides.items():
                    if self.token.value.startswith(name):
                        if self.token.value.endswith("gon"):
                            self.next()
                            shape_object = self.shape_polygon(shape, sides)
                            self.add_shape(parent, shape_object, shape)
                            return
                        elif self.token.value.endswith("gram"):
                            self.next()
                            shape_object = self.shape_star(shape, sides)
                            self.add_shape(parent, shape_object, shape)
                            return

            if self.check_words("triangle"):
                self.next()
                shape_object = self.shape_polygon(shape, 3)
                self.add_shape(parent, shape_object, shape)
                return

            sides = self.integer(None, False)
            if sides is not None:
                if self.check_words("sided", "pointed"):
                    self.next()
                    if self.check_words("polygon", "star"):
                        shape_type = self.token.value
                        function = getattr(self, "shape_" + shape_type)
                        self.next()
                        shape_object = function(shape, sides)
                        self.add_shape(parent, shape_object, shape)
                        return
                    else:
                        self.warn("Expected 'star' or 'polygon'")
                        return
                else:
                    continue

            for svg_shape in self.svg_shapes:
                if svg_shape.match(parent, self, shape):
                    return

            if not ok:
                self.next()
                break

        self.warn("Expected shape")

    def size_qualifier(self):
        base = 1

        while True:
            if self.check_words("very", "much"):
                self.next()
                base *= 1.33
            elif self.check_words("extremely"):
                self.next()
                base *= 1.5
            elif self.check_words("incredibly"):
                self.next()
                base *= 2
            else:
                break

        return base

    def size_multiplitier(self):
        lexind = self.lexer.save()
        base = self.size_qualifier()

        if self.check_words("small"):
            self.next()
            return 0.8 / base
        elif self.check_words("large", "big"):
            self.next()
            return 1.2 * base
        elif self.check_words("tiny"):
            self.next()
            return 0.5 / base
        elif self.check_words("huge"):
            self.next()
            return 1.6 * base
        else:
            self.lexer.restore(lexind)
            return None

    def add_shape(self, parent, shape_object, shape_data):
        g = shapes.Group()
        g.add_shape(shape_object)

        if shape_data.stroke and shape_data.stroke_on_top:
            g.add_shape(shape_data.stroke)

        if shape_data.color:
            fill = shapes.Fill(shape_data.color)
            g.add_shape(fill)
            shape_data.define_property("color", AnimatableType.Color, [fill.color])

        if shape_data.stroke and not shape_data.stroke_on_top:
            g.add_shape(shape_data.stroke)

        if shape_data.opacity != 1:
            g.transform.opacity.value = 100 * shape_data.opacity

        if "position" in shape_data.properties:
            center = shape_data.properties["position"].get_initial()
        else:
            center = g.bounding_box(0).center()
        g.transform.position.value = self.position(g, 0) + center
        g.transform.anchor_point.value = NVector(*center)
        shape_data.define_property("position", AnimatableType.Position, [g.transform.position], ["moves"])
        shape_data.define_property("rotation", AnimatableType.Angle, [g.transform.rotation], ["rotates"])

        if self.check_words("rotated"):
            self.next()
            g.transform.rotation.value = self.angle(0)

        parent.insert_shape(0, g)

        if self.skip_words("that"):
            self.animated_properties(shape_data)

        return g

    def position(self, shape: shapes.Group, time: float):
        px = 0
        py = 0

        qual = self.size_qualifier()

        if self.check_words("to", "in", "towards", "on", "at"):
            self.next()
            if self.check_words("the"):
                self.next()
            if self.token.type != TokenType.Word:
                self.warn("Expected position")
                return

            while True:
                if self.check_words("left"):
                    px = -1
                elif self.check_words("right"):
                    px = 1
                elif self.check_words("top"):
                    py = -1
                elif self.check_words("bottom"):
                    py = 1
                elif self.check_words("center", "middle"):
                    pass
                else:
                    break

                self.next()

            if self.check_words("side", "corner"):
                self.next()

        if px == 0 and py == 0:
            return shape.transform.position.get_value(time)

        box = shape.bounding_box(time)
        center = box.center()
        left = box.width / 2
        right = self.lottie.width - box.width / 2
        top = box.height / 2
        bottom = self.lottie.height - box.height / 2

        pos = shape.transform.position.get_value(time)
        x = pos.x
        y = pos.y
        dx = dy = 0

        if px < 0:
            dx = left - center.x
        elif px > 0:
            dx = right - center.y

        if py < 0:
            dy = top - center.y
        elif py > 0:
            dy = bottom - center.y

        return NVector(
            x + dx * qual,
            y + dy * qual,
        )

    def animation_time(self, time, required):
        if self.skip_words("at"):
            if self.skip_words("the"):
                self.require_one_of("end")
                self.next()
                return self.lottie.out_point
            return self.time(time)
        if self.skip_words("after"):
            return self.time(0) + time
        if required:
            return time
        return None

    def animated_properties(self, shape_data: ShapeData):
        time = 0

        while True:
            prop_time = self.animation_time(time, False)
            changing = self.skip_words("changing", "changes")
            possesive = self.possesive()
            found_property = None
            value = None
            loop = False

            if possesive or changing:
                for property in shape_data.properties.values():
                    if self.check_word_sequence(property.name):
                        found_property = property
                        if possesive and not changing:
                            if self.skip_words("loops"):
                                self.skip_words("back")
                                loop = True
                                break
                            else:
                                self.skip_words("changes")
                        value = self.animated_property(shape_data, property, time)
                        break
            else:
                for property in shape_data.properties.values():
                    if property.phrase and self.check_word_sequence(property.phrase):
                        found_property = property
                        value = self.animated_property(shape_data, property, time)
                        break

            if not found_property:
                self.warn("Unknown property")
                break

            if prop_time is None:
                self.prop_time = self.animation_time(time, True)
            time = prop_time

            if loop:
                found_property.loop(time)
            else:
                if value is None:
                    break

                found_property.add_keyframe(time, value)

            cont = self.skip_and()
            cont = self.skip_words("then") or cont
            if not cont:
                break

    def animated_property_value(self, property: AnimatableProperty, time):
        if property.type == AnimatableType.Angle:
            return self.angle(None)
        elif property.type == AnimatableType.Number:
            return self.number(None)
        elif property.type == AnimatableType.Integer:
            return self.integer(None)
        elif property.type == AnimatableType.Color:
            value = self.color()
            if not value:
                self.warn("Expected color")
            return value
        elif property.type == AnimatableType.Position:
            return self.position_value(property.get_value(time))
        elif property.type == AnimatableType.Size:
            return self.vector_value()
        return None

    def animated_property(self, shape_data: ShapeData, property: AnimatableProperty, time):
        relative = False
        if not self.skip_words("to"):
            relative = self.skip_words("by")

        value = self.animated_property_value(property, time)

        if value is not None and relative:
            value += property.get_value(time)

        return value

    def vector_value(self):
        x = self.number(0)
        if self.token.type == TokenType.Punctuation and self.token.value == ",":
            self.next()
        y = self.number(0)

        return NVector(x, y)

    def position_value(self, start: NVector):

        if self.token.type == TokenType.Word:
            direction = None

            if self.check_words("left") or self.check_word_sequence("the", "left"):
                direction = NVector(-1, 0)
            elif self.check_words("right") or self.check_word_sequence("the", "right"):
                direction = NVector(1, 0)
            elif self.check_words("up", "upwards", "upward"):
                direction = NVector(0, -1)
            elif self.check_words("down", "downwards", "downward"):
                direction = NVector(0, 1)

            if direction is None:
                self.warn("Expected direction or position")
                return start

            self.skip_words("by")

            return start + direction * self.number()

        return self.vector_value()

    def shape_common_property(self, shape_data: ShapeData):
        color = NVector(0, 0, 0)
        width = 4

        while True:
            got_color = self.color()
            if got_color:
                color = got_color
                continue

            quant = self.size_qualifier()
            if self.check_words("thick"):
                self.next()
                width *= 1.5 * quant
                continue
            elif self.check_words("thin"):
                self.next()
                width *= 0.6 / quant
                continue
            else:
                break

        if self.check_words("stroke", "border", "outline", "edge", "borders", "edges"):
            self.next()
            shape_data.stroke = shapes.Stroke(color, width)
            return True

        return False

    def animated_properties_callback(self, shape_data: ShapeData):
        for property in shape_data.properties.values():
            if self.check_word_sequence(property.name):
                self.skip_words("to", "is", "are")
                property.set_initial(self.animated_property_value(property, 0))
                return True
        return False

    def shape_square(self, shape_data: ShapeData):
        pos = NVector(self.lottie.width / 2, self.lottie.height / 2)
        size = NVector(shape_data.extent, shape_data.extent)
        round_base = shape_data.extent / 5
        shape = shapes.Rect(pos, size, shape_data.roundness * round_base)
        shape_data.define_property("position", AnimatableType.Position, [shape.position])
        shape_data.define_property("size", AnimatableType.Size, [shape.position])
        self.properties(shape_data, self.animated_properties_callback, [shape_data], ["with", "of"])
        return shape

    def shape_circle(self, shape_data: ShapeData):
        pos = NVector(self.lottie.width / 2, self.lottie.height / 2)
        size = NVector(shape_data.extent, shape_data.extent)
        shape = shapes.Ellipse(pos, size)
        shape_data.define_property("position", AnimatableType.Position, [shape.position])
        shape_data.define_property("size", AnimatableType.Size, [shape.position])
        self.properties(shape_data, self.animated_properties_callback, [shape_data], ["with", "of"])
        return shape

    def shape_star(self, shape_data: ShapeData, sides: int = None):
        pos = NVector(self.lottie.width / 2, self.lottie.height / 2)
        round_base = shape_data.extent / 5
        roundness = shape_data.roundness * round_base
        shape = shapes.Star()
        shape.position.value = pos
        shape.inner_roundness.value = roundness
        shape.outer_radius.value = shape_data.extent / 2
        shape.outer_roundness.value = roundness
        shape.star_type = shapes.StarType.Star
        shape.points.value = sides or 5

        def callback():
            if self.animated_properties_callback(shape_data):
                return True

            if self.check_words("diameter"):
                shape.outer_radius.value = self.number(shape.outer_radius.value) / 2
                return True
            elif self.token.type == TokenType.Number:
                if sides:
                    self.warn("Number of sides already specified")
                shape.points.value = self.integer(shape.points.value)
                if self.require_one_of("points", "sides", "point", "side"):
                    self.next()
                return True
            return False

        shape.inner_radius.value = shape.outer_radius.value / 2

        shape_data.define_property("position", AnimatableType.Position, [shape.position])
        shape_data.define_property("outer radius", AnimatableType.Number, [shape.outer_radius])
        shape_data.define_property("radius", AnimatableType.Number, [shape.outer_radius])
        shape_data.define_property("inner radius", AnimatableType.Number, [shape.inner_radius])
        self.properties(shape_data, callback, [], ["with", "of", "has"])

        return shape

    def shape_polygon(self, shape_data: ShapeData, sides: int = None):
        shape = self.shape_star(shape_data, sides)
        shape.inner_radius = None
        shape.inner_roundness = None
        shape.star_type = shapes.StarType.Polygon
        return shape

    def angle_direction(self):
        if self.check_words("clockwise"):
            self.next()
            return 1
        elif self.check_words("counter"):
            self.next()
            if self.check_words("clockwise"):
                self.next()
            return -1
        return 0

    def fraction(self):
        if self.article():
            amount = 1
        else:
            amount = self.number(1)

        if self.check_words("full", "entire"):
            self.next()
            return amount, True
        elif self.check_words("half", "halfs"):
            self.next()
            return amount / 2, True
        elif self.check_words("third", "thirds"):
            self.next()
            return amount / 3, True
        elif self.check_words("quarter", "quarters"):
            self.next()
            return amount / 3, True

        return amount, False

    def angle(self, default):
        direction = self.angle_direction()

        amount, has_fraction = self.fraction()

        if self.skip_and():
            more_frac = self.fraction()[0]
            if has_fraction:
                amount += amount * more_frac
            else:
                amount += more_frac

        if self.check_words("turns"):
            self.next()
            amount *= 360
        elif self.require_one_of("degrees"):
            self.next()
        elif self.check_word_sequence(["pi", "radians"]):
            amount *= 180
            if direction == 0:
                direction = -1

        if direction == 0:
            direction = 1
        return amount * direction

    def rect_properties(self, shape_data: ShapeData, shape):
        extent = shape_data.extent
        parse_data = {
            "width": None,
            "height": None,
            "ratio": math.sqrt(2),
            "size_specified": False
        }
        handle_orientation = True

        shape_data.define_property("position", AnimatableType.Position, [shape.position])
        shape_data.define_property("size", AnimatableType.Size, [shape.size])

        def callback():
            if self.animated_properties_callback(shape_data):
                return True

            if self.check_words("ratio"):
                self.next()
                ratio = self.fraction()[0]
                if ratio <= 0:
                    self.warn("Ratio must be positive")
                else:
                    parse_data["ratio"] = ratio
                return True
            elif self.check_words("width"):
                self.next()
                parse_data["width"] = self.number(0)
                return True
            elif self.check_words("height"):
                self.next()
                parse_data["height"] = self.number(0)
                return True
            elif self.check_words("size"):
                parse_data["size_specified"] = True

            return False

        self.properties(shape_data, callback, [], ["with", "of", "has"])

        if not parse_data["size_specified"]:
            width = parse_data["width"]
            height = parse_data["height"]
            ratio = parse_data["ratio"]

            if width is None and height is None:
                width = extent
                height = width / ratio
            elif width is None:
                width = height * ratio
            elif height is None:
                height = width / ratio
            else:
                handle_orientation = False

            if handle_orientation:
                if (width > height and shape_data.portrait) or (width < height and not shape_data.portrait):
                    width, height = height, width

            shape.size.value = NVector(width, height)

    def shape_rectangle(self, shape_data: ShapeData):
        pos = NVector(self.lottie.width / 2, self.lottie.height / 2)
        round_base = shape_data.extent / 5

        shape = shapes.Rect(pos, NVector(0, 0), shape_data.roundness * round_base)
        shape_data.define_property("roundness", AnimatableType.Number, [shape.rounded])
        self.rect_properties(shape_data, shape)
        return shape

    def shape_ellipse(self, shape_data: ShapeData):
        pos = NVector(self.lottie.width / 2, self.lottie.height / 2)
        shape = shapes.Ellipse(pos, NVector(0, 0))
        self.rect_properties(shape_data, shape)
        return shape

    def shape_text(self, shape_data: ShapeData):
        text = self.string("")
        font = self.font.clone()
        shape = font.render(text)

        box = shape.bounding_box()

        center = box.center()
        anim_center = NVector(self.lottie.width / 2, self.lottie.height / 2)
        shape.transform.anchor_point.value = anim_center
        shape.transform.position.value.y = self.lottie.height - center.y
        shape.transform.position.value.x = self.lottie.width - center.x

        scale = shape_data.size_multiplitier
        if box.width > self.lottie.width > 0:
            scale *= self.lottie.width / box.width

        if scale != 1:
            wrapper = shapes.Group()
            wrapper.transform.position.value = wrapper.transform.anchor_point.value = anim_center
            wrapper.transform.scale.value *= scale
            wrapper.add_shape(shape)
            return wrapper

        return shape


class SvgLoader:
    def __init__(self):
        self.cache = {}

    def load(self, filename):
        if filename in self.cache:
            return self.cache[filename]

        anim = parse_svg_file(filename)
        self.cache[filename] = anim
        return anim


class SvgFeature:
    def __init__(self, layer_names, colors):
        self.layer_names = layer_names
        self.colors = [parse_color(color) for color in colors]

    def iter_stylers(self, lottie_object):
        objects = []
        if self.layer_names:
            for layer_name in self.layer_names:
                found = lottie_object.find(layer_name)
                if found:
                    objects.append(found)
        else:
            objects = [lottie_object]

        for object in objects:
            for styler in object.find_all((shapes.Fill, shapes.Stroke)):
                if len(self.colors) == 0 or styler.color.value in self.colors:
                    yield styler

    def process(self, lottie_object, new_color):
        for styler in self.iter_stylers(lottie_object):
            styler.color.value = new_color


class SvgShape:
    def __init__(self, file, phrase, feature_map, main_feature, facing_direction, svg_loader: SvgLoader):
        self.file = file
        self.phrase = phrase
        self.feature_map = feature_map
        if isinstance(main_feature, str):
            self.main_feature = feature_map[main_feature]
        else:
            self.main_feature = main_feature
        self.facing_direction = facing_direction
        self.svg_loader = svg_loader

    def callback(self, parser: Parser, shape_data: ShapeData):
        lexind = parser.lexer.save()
        color = parser.color()
        if color:
            if parser.check_words(*self.feature_map.keys()):
                feature_name = parser.lexer.token.value
                shape_data.properties[feature_name].set_initial(color)
                parser.next()
                return True
        else:
            parser.lexer.restore(lexind)

        return False

    def match(self, parent, parser: Parser, shape_data: ShapeData):
        if not parser.check_word_sequence(self.phrase):
            return False

        shape_data.stroke_on_top = False
        svg_anim = parse_svg_file(self.file, 0, parser.lottie.out_point, parser.lottie.frame_rate)
        layer = svg_anim.layers[0].clone()
        group = shapes.Group()
        group.name = layer.name
        group.shapes = layer.shapes + group.shapes
        layer.transform.clone_into(group.transform)

        wrapper = shapes.Group()
        wrapper.add_shape(group)

        delta_ap = group.bounding_box(0).center()
        wrapper.transform.anchor_point.value += delta_ap
        wrapper.transform.position.value += delta_ap
        wrapper.transform.scale.value *= shape_data.size_multiplitier

        for name, feature in self.feature_map.items():
            singular = name[:-1] if name.endswith("s") else name
            shape_data.properties[name] = AnimatableProperty(singular + " color", None, [], AnimatableType.Color)
            for styler in feature.iter_stylers(group):
                shape_data.add_property(name, styler.color)

        if self.main_feature:
            shape_data.define_property("color", AnimatableType.Color, [])

            for styler in self.main_feature.iter_stylers(group):
                shape_data.add_property("color", styler.color)

            if shape_data.color and shape_data.color_explicit:
                shape_data.properties["color"].set_initial(shape_data.color)

        parser.properties(shape_data, self.callback, [parser, shape_data], ["with"])

        if self.facing_direction != 0 and parser.check_words("facing", "looking"):
            parser.next()
            if parser.skip_words("to"):
                parser.skip_words("the")

            if parser.check_words("left", "right"):
                direction = -1 if parser.token.value == "left" else 1
                if direction != self.facing_direction:
                    wrapper.transform.scale.value.x *= -1
            else:
                parser.warn("Missing facing direction")

            parser.next()

        shape_data.color = None
        parser.add_shape(parent, wrapper, shape_data)
        return True



import json
import zipfile

from .base import importer
from ..parsers.baseporter import ExtraOption
from ..parsers.tgs import parse_tgs
from ..objects import Animation, assets


@importer("dotLottie Archive", ["lottie"], [
    ExtraOption("id", help="ID of the animation to extract", default=None)
], slug="dotlottie")
def import_dotlottie(file, id=None):
    with zipfile.ZipFile(file) as zf:
        with zf.open("manifest.json") as manifest:
            meta = json.load(manifest)

        if id is None:
            id = meta["animations"][0]["id"]

        info = zf.getinfo("animations/%s.json" % id)

        with zf.open(info) as animfile:
            an = Animation.load(json.load(animfile))
            if an.assets:
                for asset in an.assets:
                    if isinstance(asset, assets.Image) and not asset.is_embedded:
                        fname = asset.path + asset.file_name
                        if fname in zf.namelist():
                            with zf.open(fname) as imgfile:
                                asset.load(imgfile)
            return an



"""!
Package with all the Lottie Python bindings
"""
from . import (
    animation, base, effects, enums, helpers, layers, shapes, assets, easing,
    text, bezier, composition, styles
)
from .animation import Animation
from .layers import *
from .shapes import *
from .styles import *
from .assets import Precomp, Image, Asset, FileAsset
from .bezier import Bezier
from .composition import Composition

__all__ = [
    "animation", "base", "effects", "enums", "helpers", "layers", "shapes", "assets",
    "easing", "text", "bezier", "styles",
    "Animation",
    "NullLayer", "TextLayer", "ShapeLayer", "ImageLayer", "PreCompLayer", "SolidColorLayer",
    "Rect", "Fill", "Trim", "Repeater", "GradientFill", "Stroke", "RoundedCorners", "Path",
    "TransformShape", "Group", "Star", "Ellipse", "Merge", "GradientStroke",
    "Bezier", "Precomp", "Composition", "Image", "Asset", "FileAsset", "LayerStyle"
]



import enum
import math
import colorsys
from ..nvector import NVector


def from_uint8(r, g, b, a=255):
    return Color(r, g, b, a) / 255


class ColorMode(enum.Enum):
    ## sRGB, Components in [0, 1]
    RGB = enum.auto()
    ## HSV, components in [0, 1]
    HSV = enum.auto()
    ## HSL, components in [0, 1]
    HSL = enum.auto()
    ## CIE XYZ with Illuminant D65. Components in [0, 1]
    XYZ = enum.auto()
    ## CIE L*u*v*
    LUV = enum.auto()
    ## CIE Lch(uv), polar version of LUV where C is the radius and H an angle in radians
    LCH_uv = enum.auto()
    ## CIE L*a*b*
    LAB = enum.auto()
    ## CIE LCh(ab), polar version of LAB where C is the radius and H an angle in radians
    #LCH_ab = enum.auto()


def _clamp(x):
    return max(0, min(1, x))


class Conversion:
    _conv_paths = {
        (ColorMode.RGB, ColorMode.RGB): [],
        (ColorMode.RGB, ColorMode.HSV): [],
        (ColorMode.RGB, ColorMode.HSL): [],
        (ColorMode.RGB, ColorMode.XYZ): [],
        (ColorMode.RGB, ColorMode.LUV): [ColorMode.XYZ],
        (ColorMode.RGB, ColorMode.LAB): [ColorMode.XYZ],
        (ColorMode.RGB, ColorMode.LCH_uv): [ColorMode.XYZ, ColorMode.LUV],
        #(ColorMode.RGB, ColorMode.LCH_ab): [ColorMode.XYZ, ColorMode.LAB],

        (ColorMode.HSV, ColorMode.RGB): [],
        (ColorMode.HSV, ColorMode.HSV): [],
        (ColorMode.HSV, ColorMode.HSL): [],
        (ColorMode.HSV, ColorMode.XYZ): [ColorMode.RGB],
        (ColorMode.HSV, ColorMode.LUV): [ColorMode.RGB, ColorMode.XYZ],
        (ColorMode.HSV, ColorMode.LAB): [ColorMode.RGB, ColorMode.XYZ],
        (ColorMode.HSV, ColorMode.LCH_uv): [ColorMode.RGB, ColorMode.XYZ, ColorMode.LUV],
        #(ColorMode.HSV, ColorMode.LCH_ab): [ColorMode.RGB, ColorMode.XYZ, ColorMode.LAB],

        (ColorMode.HSL, ColorMode.RGB): [],
        (ColorMode.HSL, ColorMode.HSV): [],
        (ColorMode.HSL, ColorMode.HSL): [],
        (ColorMode.HSL, ColorMode.XYZ): [ColorMode.RGB],
        (ColorMode.HSL, ColorMode.LUV): [ColorMode.RGB, ColorMode.XYZ],
        (ColorMode.HSL, ColorMode.LAB): [ColorMode.RGB, ColorMode.XYZ],
        (ColorMode.HSL, ColorMode.LCH_uv): [ColorMode.RGB, ColorMode.XYZ, ColorMode.LUV],
        #(ColorMode.HSL, ColorMode.LCH_ab): [ColorMode.RGB, ColorMode.XYZ, ColorMode.LAB],

        (ColorMode.XYZ, ColorMode.RGB): [],
        (ColorMode.XYZ, ColorMode.HSV): [ColorMode.RGB],
        (ColorMode.XYZ, ColorMode.HSL): [ColorMode.RGB],
        (ColorMode.XYZ, ColorMode.XYZ): [],
        (ColorMode.XYZ, ColorMode.LUV): [],
        (ColorMode.XYZ, ColorMode.LAB): [],
        (ColorMode.XYZ, ColorMode.LCH_uv): [ColorMode.LUV],
        #(ColorMode.XYZ, ColorMode.LCH_ab): [ColorMode.LAB],

        (ColorMode.LCH_uv, ColorMode.RGB): [ColorMode.LUV, ColorMode.XYZ],
        (ColorMode.LCH_uv, ColorMode.HSV): [ColorMode.LUV, ColorMode.XYZ, ColorMode.RGB],
        (ColorMode.LCH_uv, ColorMode.HSL): [ColorMode.LUV, ColorMode.XYZ, ColorMode.RGB],
        (ColorMode.LCH_uv, ColorMode.XYZ): [ColorMode.LUV],
        (ColorMode.LCH_uv, ColorMode.LUV): [],
        (ColorMode.LCH_uv, ColorMode.LAB): [ColorMode.LUV, ColorMode.XYZ],
        (ColorMode.LCH_uv, ColorMode.LCH_uv): [],
        #(ColorMode.LCH_uv, ColorMode.LCH_ab): [ColorMode.LUV, ColorMode.XYZ, ColorMode.LAB],

        (ColorMode.LUV, ColorMode.RGB): [ColorMode.XYZ],
        (ColorMode.LUV, ColorMode.HSV): [ColorMode.XYZ, ColorMode.RGB],
        (ColorMode.LUV, ColorMode.HSL): [ColorMode.XYZ, ColorMode.RGB],
        (ColorMode.LUV, ColorMode.XYZ): [],
        (ColorMode.LUV, ColorMode.LUV): [],
        (ColorMode.LUV, ColorMode.LAB): [ColorMode.XYZ],
        (ColorMode.LUV, ColorMode.LCH_uv): [],
        #(ColorMode.LUV, ColorMode.LCH_ab): [ColorMode.XYZ, ColorMode.LAB],

        (ColorMode.LAB, ColorMode.RGB): [ColorMode.XYZ],
        (ColorMode.LAB, ColorMode.HSV): [ColorMode.XYZ, ColorMode.RGB],
        (ColorMode.LAB, ColorMode.HSL): [ColorMode.XYZ, ColorMode.RGB],
        (ColorMode.LAB, ColorMode.XYZ): [],
        (ColorMode.LAB, ColorMode.LUV): [ColorMode.XYZ],
        (ColorMode.LAB, ColorMode.LAB): [],
        (ColorMode.LAB, ColorMode.LCH_uv): [ColorMode.XYZ, ColorMode.LUV],
        #(ColorMode.LAB, ColorMode.LCH_ab): [],

        #(ColorMode.LCH_ab, ColorMode.RGB): [ColorMode.LAB, ColorMode.XYZ],
        #(ColorMode.LCH_ab, ColorMode.HSV): [ColorMode.LAB, ColorMode.XYZ, ColorMode.RGB],
        #(ColorMode.LCH_ab, ColorMode.HSL): [ColorMode.LAB, ColorMode.XYZ, ColorMode.RGB],
        #(ColorMode.LCH_ab, ColorMode.XYZ): [ColorMode.LAB],
        #(ColorMode.LCH_ab, ColorMode.LUV): [ColorMode.LAB, ColorMode.XYZ],
        #(ColorMode.LCH_ab, ColorMode.LAB): [],
        #(ColorMode.LCH_ab, ColorMode.LCH_uv): [ColorMode.LAB, ColorMode.XYZ, ColorMode.LUV],
        #(ColorMode.LCH_ab, ColorMode.LCH_ab): [],
    }

    @staticmethod
    def rgb_to_hsv(r, g, b):
        return colorsys.rgb_to_hsv(r, g, b)

    @staticmethod
    def hsv_to_rgb(h, s, v):
        return colorsys.hsv_to_rgb(h, s, v)

    @staticmethod
    def hsl_to_hsv(h, s_hsl, l):
        v = l + s_hsl * min(l, 1 - l)
        s_hsv = 0 if v == 0 else 2 - 2 * l / v
        return (h, s_hsv, v)

    @staticmethod
    def hsv_to_hsl(h, s_hsv, v):
        l = v - v * s_hsv / 2
        s_hsl = 0 if l in (0, 1) else (v - l) / min(l, 1 - l)
        return (h, s_hsl, l)

    @staticmethod
    def rgb_to_hsl(r, g, b):
        h, l, s = colorsys.rgb_to_hls(r, g, b)
        return (h, s, l)

    @staticmethod
    def hsl_to_rgb(h, s, l):
        return colorsys.hls_to_rgb(h, l, s)

    # http://w3.uqo.ca/missaoui/Publications/TRColorSpace.zip
    #@staticmethod
    #def rgb_to_hcl(r, g, b, gamma=3, y0=100):
        #maxc = max(r, g, b)
        #minc = min(r, g, b)
        #if maxc > 0:
            #alpha = 1/y0 * minc / maxc
        #else:
            #alpha = 0
        #q = math.e ** (alpha * gamma)
        #h = math.atan2(g - b, r - g)
        #if h < 0:
            #h += 2*math.pi
        #h /= 2*math.pi
        #c = q / 3 * (abs(r-g) + abs(g-b) + abs(b-r))
        #l = (q * maxc + (q-1) * minc) / 2
        #return (h, c, l)

    #@staticmethod
    #def hcl_to_rgb(h, c, l, gamma=3, y0=100):
        #h *= 2*math.pi

        #q = math.e ** ((1 - 2*c / 4*l) * gamma / y0)
        #minc = (4*l - 3*c) / (4*q - 2)
        #maxc = minc + 3*c / 2*q

        #if h <= math.pi * 1 / 3:
            #tan = math.tan(3/2*h)
            #r = maxc
            #b = minc
            #g = (r * tan + b) / (1 + tan)
        #elif h <= math.pi * 2 / 3:
            #tan = math.tan(3/4*(h-math.pi))
            #g = maxc
            #b = minc
            #r = (g * (1+tan) - b) / tan
        #elif h <= math.pi * 3 / 3:
            #tan = math.tan(3/4*(h-math.pi))
            #g = maxc
            #r = minc
            #b = g * (1+tan) - r * tan
        #elif h <= math.pi * 4 / 3:
            #tan = math.tan(3/2*(h+math.pi))
            #b = maxc
            #r = minc
            #g = (r * tan + b) / (1 + tan)
        #elif h <= math.pi * 5 / 3:
            #tan = math.tan(3/4*h)
            #b = maxc
            #g = minc
            #r = (g * (1+tan) - b) / tan
        #else:
            #tan = math.tan(3/4*h)
            #r = maxc
            #g = minc
            #b = g * (1+tan) - r * tan

        #return _clamp(r), _clamp(g), _clamp(b)

    @staticmethod
    def rgb_to_xyz(r, g, b):
        def _gamma(v):
            return v / 12.92 if v <= 0.04045 else ((v + 0.055) / 1.055) ** 2.4
        rgb = (_gamma(r), _gamma(g), _gamma(b))
        matrix = [
            [0.4124564, 0.3575761, 0.1804375],
            [0.2126729, 0.7151522, 0.0721750],
            [0.0193339, 0.1191920, 0.9503041],
        ]
        return tuple(
            sum(rgb[i] * c for i, c in enumerate(row))
            for row in matrix
        )

    @staticmethod
    def xyz_to_rgb(x, y, z):
        def _gamma1(v):
            return _clamp(v * 12.92 if v <= 0.0031308 else v ** (1/2.4) * 1.055 - 0.055)
        matrix = [
            [+3.2404542, -1.5371385, -0.4985314],
            [-0.9692660, +1.8760108, +0.0415560],
            [+0.0556434, -0.2040259, +1.0572252],
        ]
        xyz = (x, y, z)
        return tuple(map(_gamma1, (
            sum(xyz[i] * c for i, c in enumerate(row))
            for row in matrix
        )))

    @staticmethod
    def xyz_to_luv(x, y, z):
        u1r = 0.2009
        v1r = 0.4610
        yr = 100

        kap = (29/3)**3
        eps = (6/29)**3

        try:
            u1 = 4*x / (x + 15*y + 3*z)
            v1 = 9*y / (x + 15*y + 3*z)
        except ZeroDivisionError:
            return 0, 0, 0

        y_r = y/yr
        l = 166 * y_r ** (1/3) - 16 if y_r > eps else kap * y_r
        u = 13 * l * (u1 - u1r)
        v = 13 * l * (v1 - v1r)
        return l, u, v

    @staticmethod
    def luv_to_xyz(l, u, v):
        u1r = 0.2009
        v1r = 0.4610
        yr = 100

        kap = (29/3)**3

        if l == 0:
            u1 = u1r
            v1 = v1r
        else:
            u1 = u / (13 * l) + u1r
            v1 = v / (13 * l) + v1r

        y = yr * l / kap if l <= 8 else yr * ((l + 16) / 116) ** 3
        x = y * 9*u1 / (4*v1)
        z = y * (12 - 3*u1 - 20*v1) / (4*v1)
        return x, y, z

    @staticmethod
    def luv_to_lch_uv(l, u, v):
        c = math.hypot(u, v)
        h = math.atan2(v, u)
        if h < 0:
            h += math.tau
        return l, c, h

    @staticmethod
    def lch_uv_to_luv(l, c, h):
        u = math.cos(h) * c
        v = math.sin(h) * c
        return l, u, v

    @staticmethod
    def xyz_to_lab(x, y, z):
        # D65 Illuminant aka sRGB(1,1,1)
        xn = 0.950489
        yn = 1
        zn = 108.8840

        delta = 6 / 29

        def f(t):
            return t ** (1/3) if t > delta ** 3 else t / (3*delta**2) + 4/29

        fy = f(y/yn)
        l = 116 * fy - 16
        a = 500 * (f(x/xn) - fy)
        b = 200 * (fy - f(z/zn))

        return l, a, b

    @staticmethod
    def lab_to_xyz(l, a, b):
        # D65 Illuminant aka sRGB(1,1,1)
        xn = 0.950489
        yn = 1
        zn = 108.8840

        delta = 6 / 29

        def f1(t):
            return t**3 if t > delta else 3*delta**2*(t-4/29)

        l1 = (l+16) / 116
        x = xn * f1(l1+a/500)
        y = yn * f1(l1)
        z = zn * f1(l1-b/200)

        return x, y, z

    #@staticmethod
    #def lab_to_lch_ab(l, a, b):
        #c = math.hypot(a, b)
        #h = math.atan2(b, a)
        #if h < 0:
            #h += math.tau
        #return l, c, h

    #@staticmethod
    #def lch_ab_to_lab(l, c, h):
        #a = math.cos(h) * c
        #b = math.sin(h) * c
        #return l, a, b

    @staticmethod
    def conv_func(mode_from, mode_to):
        return getattr(Conversion, "%s_to_%s" % (mode_from.name.lower(), mode_to.name.lower()), None)

    @staticmethod
    def convert(tuple, mode_from, mode_to):
        if mode_from == mode_to:
            return tuple

        if len(tuple) == 4:
            alpha = tuple[3]
            tuple = tuple[:3]
        else:
            alpha = None

        func = Conversion.conv_func(mode_from, mode_to)
        if func:
            return func(*tuple)

        if (mode_from, mode_to) in Conversion._conv_paths:
            steps = Conversion._conv_paths[(mode_from, mode_to)] + [mode_to]
            for step in steps:
                func = Conversion.conv_func(mode_from, step)
                if not func:
                    raise ValueError("Missing definition for conversion from %s to %s" % (mode_from, step))
                tuple = func(*tuple)
                mode_from = step
            if alpha is not None:
                tuple += (alpha,)
            return tuple

        raise ValueError("No conversion path from %s to %s" % (mode_from, mode_to))


class Color(NVector):
    Mode = ColorMode

    def __init__(self, c1=0, c2=0, c3=0, a=1, *, mode=ColorMode.RGB):
        if isinstance(a, ColorMode):
            raise TypeError("Please update the Color constructor")
        super().__init__(c1, c2, c3, a)
        self._mode = mode

    @property
    def mode(self):
        return self._mode

    def convert(self, v):
        if v == self._mode:
            return self

        self.components = list(Conversion.convert(self.components, self._mode, v))

        self._mode = v
        return self

    def clone(self):
        return Color(*self.components, mode=self._mode)

    def converted(self, mode):
        return self.clone().convert(mode)

    def to_rgb(self):
        return self.converted(ColorMode.RGB)

    def __repr__(self):
        return "<%s %s [%.3f, %.3f, %.3f, %.3f]>" % (
            (self.__class__.__name__, self.mode.name) + tuple(self.components)
        )

    def component_names(self):
        comps = None

        if self._mode == ColorMode.RGB:
            comps = ({"r", "red"}, {"g", "green"}, {"b", "blue"})
        elif self._mode == ColorMode.HSV:
            comps = ({"h", "hue"}, {"s", "saturation"}, {"v", "value"})
        elif self._mode == ColorMode.HSL:
            comps = ({"h", "hue"}, {"s", "saturation"}, {"l", "lightness"})
        elif self._mode == ColorMode.LCH_uv:  # in (ColorMode.LCH_uv, ColorMode.LCH_ab):
            comps = ({"l", "luma", "luminance"}, {"c", "choma"}, {"h", "hue"})
        elif self._mode == ColorMode.XYZ:
            comps = "xyz"
        elif self._mode == ColorMode.LUV:
            comps = "luv"
        elif self._mode == ColorMode.LAB:
            comps = "lab"

        return comps

    def _attrindex(self, name):
        comps = self.component_names()

        if comps:
            for i, vals in enumerate(comps):
                if name in vals:
                    return i

        return None

    def __getattr__(self, name):
        if name not in vars(self) and name not in {"_mode", "components"}:
            i = self._attrindex(name)
            if i is not None:
                return self.components[i]
        raise AttributeError(name)

    def __setattr__(self, name, value):
        if name not in vars(self) and name not in {"_mode", "components"}:
            i = self._attrindex(name)
            if i is not None:
                self.components[i] = value
                return
        return super().__setattr__(name, value)


def color_to_hex(color: Color):
    conv = color.to_rgb()
    return "#%02x%02x%02x" % tuple(map(int, (conv * 255).components[:3]))


def color_from_hex(hex: str):
    from ..parsers.svg.importer import parse_color
    return parse_color(hex)



import math
from xml.dom import minidom

from ... import objects
from ...nvector import NVector
from ...utils import restructure
from . import api, ast


blend_modes = {
    objects.BlendMode.Normal: api.BlendMethod.Composite,
    objects.BlendMode.Multiply: api.BlendMethod.Multiply,
    objects.BlendMode.Screen: api.BlendMethod.Screen,
    objects.BlendMode.Overlay: api.BlendMethod.Overlay,
    objects.BlendMode.Darken: api.BlendMethod.Darken,
    objects.BlendMode.Lighten: api.BlendMethod.Lighten,
    objects.BlendMode.HardLight: api.BlendMethod.HardLight,
    objects.BlendMode.Difference: api.BlendMethod.Difference,
    objects.BlendMode.Hue: api.BlendMethod.Hue,
    objects.BlendMode.Saturation: api.BlendMethod.Saturation,
    objects.BlendMode.Color: api.BlendMethod.Color,
    objects.BlendMode.Luminosity: api.BlendMethod.Luminosity,
    objects.BlendMode.Exclusion: api.BlendMethod.Difference,
    objects.BlendMode.SoftLight: api.BlendMethod.Multiply,
    objects.BlendMode.ColorDodge: api.BlendMethod.Composite,
    objects.BlendMode.ColorBurn: api.BlendMethod.Composite,
}


class SifBuilder(restructure.AbstractBuilder):
    def __init__(self, gamma=1.0):
        """
        @todo Add gamma option to lottie_convert.py
        """
        super().__init__()
        self.canvas = api.Canvas()
        self.canvas.version = "1.2"
        self.canvas.gamma_r = self.canvas.gamma_g = self.canvas.gamma_b = gamma
        self.autoid = objects.base.Index()

    def _on_animation(self, animation: objects.Animation):
        if animation.name:
            self.canvas.name = animation.name
        self.canvas.width = animation.width
        self.canvas.height = animation.height
        self.canvas.xres = animation.width
        self.canvas.yres = animation.height
        self.canvas.view_box = NVector(0, 0, animation.width, animation.height)
        self.canvas.fps = animation.frame_rate
        self.canvas.begin_time = api.FrameTime.frame(animation.in_point)
        self.canvas.end_time = api.FrameTime.frame(animation.out_point)
        self.canvas.antialias = True
        return self.canvas

    def _on_precomp(self, id, dom_parent, layers):
        g = dom_parent.add_layer(api.GroupLayer())
        g.desc = id
        for layer_builder in layers:
            self.process_layer(layer_builder, g)

    def _on_layer(self, layer_builder, dom_parent):
        layer = self.layer_from_lottie(api.GroupLayer, layer_builder.lottie, dom_parent)
        if not layer_builder.lottie.name:
            layer.desc = layer_builder.lottie.__class__.__name__

        bm = getattr(layer_builder.lottie, "blend_mode", None)
        if bm is None:
            bm = objects.BlendMode.Normal
        layer.blend_method = blend_modes[bm]

        layer.time_drilation = getattr(layer_builder.lottie, "stretch", 1) or 1

        in_point = getattr(layer_builder.lottie, "in_point", 0)
        layer.time_offset.value = api.FrameTime.frame(in_point)

        #layer.canvas.end_time = api.FrameTime.frame(out_point)
        return layer

    def layer_from_lottie(self, type, lottie, dom_parent):
        g = dom_parent.add_layer(type())
        if lottie.name:
            g.desc = lottie.name
        g.active = not lottie.hidden
        transf = getattr(lottie, "transform", None)
        if transf:
            self.set_transform(g, transf)

        if isinstance(lottie, objects.NullLayer):
            g.amount.value = 1

        return g

    def _get_scale(self, transform):
        def func(keyframe):
            t = keyframe.time if keyframe else 0
            scale_x, scale_y = transform.scale.get_value(t)[:2]
            scale_x /= 100
            scale_y /= 100
            skew = transform.skew.get_value(t) if transform.skew else 0
            c = math.cos(skew * math.pi / 180)
            if c != 0:
                scale_y *= 1 / c
            return NVector(scale_x, scale_y)
        return func

    def set_transform(self, group, transform):
        composite = group.transformation

        if transform.position:
            composite.offset = self.process_vector(transform.position)

        if transform.scale:
            keyframes = self._merge_keyframes([transform.scale, transform.skew])
            composite.scale = self.process_vector_ext(keyframes, self._get_scale(transform))

        composite.skew_angle = self.process_scalar(transform.skew or objects.Value(0))

        if transform.rotation:
            composite.angle = self.process_scalar(transform.rotation)

        if transform.opacity:
            group.amount = self.process_scalar(transform.opacity, 1/100)

        if transform.anchor_point:
            group.origin = self.process_vector(transform.anchor_point)

        # TODO get z_depth from position
        composite.z_depth = 0

    def process_vector(self, multidim):
        def getter(keyframe):
            if keyframe is None:
                v = multidim.value
            else:
                v = keyframe.start
            return NVector(v[0], v[1])

        return self.process_vector_ext(multidim.keyframes, getter)

    def process_vector_ext(self, kframes, getter):
        if kframes is not None:
            wrap = ast.SifAnimated()
            for i in range(len(kframes)):
                keyframe = kframes[i]
                waypoint = wrap.add_keyframe(getter(keyframe), api.FrameTime.frame(keyframe.time))

                if i > 0:
                    prev = kframes[i-1]
                    if prev.hold:
                        waypoint.before = api.Interpolation.Constant
                    elif prev.in_value and prev.in_value.x < 1:
                        waypoint.before = api.Interpolation.Ease
                    else:
                        waypoint.before = api.Interpolation.Linear
                else:
                    waypoint.before = api.Interpolation.Linear

                if keyframe.hold:
                    waypoint.after = api.Interpolation.Constant
                elif keyframe.out_value and keyframe.out_value.x > 0:
                    waypoint.after = api.Interpolation.Ease
                else:
                    waypoint.after = api.Interpolation.Linear
        else:
            wrap = api.SifValue(getter(None))

        return wrap

    def process_scalar(self, value, mult=None):
        def getter(keyframe):
            if keyframe is None:
                v = value.value
            else:
                v = keyframe.start[0]
            if mult is not None:
                v *= mult
            return v
        return self.process_vector_ext(value.keyframes, getter)

    def _on_shape(self, shape, group, dom_parent):
        layers = []
        if not hasattr(shape, "to_bezier"):
            return []

        if group.stroke:
            sif_shape = self.build_path(api.OutlineLayer, shape.to_bezier(), dom_parent, shape)
            self.apply_group_stroke(sif_shape, group.stroke)
            layers.append(sif_shape)

        if group.fill:
            sif_shape = self.build_path(api.RegionLayer, shape.to_bezier(), dom_parent, shape)
            layers.append(sif_shape)
            self.apply_group_fill(sif_shape, group.fill)

        return layers

    def _merge_keyframes(self, props):
        keyframes = {}
        for prop in props:
            if prop is not None and prop.animated:
                keyframes.update({kf.time: kf for kf in prop.keyframes})
        return list(sorted(keyframes.values(), key=lambda kf: kf.time)) or None

    def apply_origin(self, sif_shape, lottie_shape):
        if hasattr(lottie_shape, "position"):
            sif_shape.origin.value = lottie_shape.position.get_value()
        else:
            sif_shape.origin.value = lottie_shape.bounding_box().center()

    def apply_group_fill(self, sif_shape, fill):
        ## @todo gradients?
        if hasattr(fill, "colors"):
            return

        def getter(keyframe):
            if keyframe is None:
                v = fill.color.value
            else:
                v = keyframe.start
            return self.canvas.make_color(*v)

        sif_shape.color = self.process_vector_ext(fill.color.keyframes, getter)

        def get_op(keyframe):
            if keyframe is None:
                v = fill.opacity.value
            else:
                v = keyframe.start[0]
            v /= 100
            return v

        sif_shape.amount = self.process_vector_ext(fill.opacity.keyframes, get_op)

    def apply_group_stroke(self, sif_shape, stroke):
        self.apply_group_fill(sif_shape, stroke)
        sif_shape.sharp_cusps.value = stroke.line_join == objects.LineJoin.Miter
        round_cap = stroke.line_cap == objects.LineCap.Round
        sif_shape.round_tip_0.value = round_cap
        sif_shape.round_tip_1.value = round_cap
        sif_shape.width = self.process_scalar(stroke.width, 0.5)

    def build_path(self, type, path, dom_parent, lottie_shape):
        layer = self.layer_from_lottie(type, lottie_shape, dom_parent)
        self.apply_origin(layer, lottie_shape)
        startbez = path.shape.get_value()
        layer.bline.loop = startbez.closed
        nverts = len(startbez.vertices)
        for point in range(nverts):
            self.bezier_point(path, point, layer.bline, layer.origin.value)
        return layer

    def bezier_point(self, lottie_path, point_index, sif_parent, offset):
        composite = api.BlinePoint()

        def get_point(keyframe):
            if keyframe is None:
                bezier = lottie_path.shape.value
            else:
                bezier = keyframe.start
            if not bezier:
                #elem.parentNode.parentNode.removeChild(elem.parentNode)
                return
            vert = bezier.vertices[point_index]
            return NVector(vert[0], vert[1]) - offset

        composite.point = self.process_vector_ext(lottie_path.shape.keyframes, get_point)
        composite.split.value = True
        composite.split_radius.value = True
        composite.split_angle.value = True

        def get_tangent(keyframe):
            if keyframe is None:
                bezier = lottie_path.shape.value
            else:
                bezier = keyframe.start
            if not bezier:
                #elem.parentNode.parentNode.removeChild(elem.parentNode)
                return

            inp = getattr(bezier, which_point)[point_index]
            return NVector(inp.x, inp.y) * 3 * mult

        mult = -1
        which_point = "in_tangents"
        composite.t1 = self.process_vector_ext(lottie_path.shape.keyframes, get_tangent)

        mult = 1
        which_point = "out_tangents"
        composite.t2 = self.process_vector_ext(lottie_path.shape.keyframes, get_tangent)
        sif_parent.points.append(composite)

    def _on_shapegroup(self, shape_group, dom_parent):
        if shape_group.empty():
            return

        layer = self.layer_from_lottie(api.GroupLayer, shape_group.lottie, dom_parent)

        self.shapegroup_process_children(shape_group, layer)

    def _modifier_inner_group(self, modifier, shapegroup, dom_parent):
        layer = dom_parent.add_layer(api.GroupLayer())
        self.shapegroup_process_child(modifier.child, shapegroup, layer)
        return layer

    def _on_shape_modifier(self, modifier, shapegroup, dom_parent):
        layer = dom_parent.add_layer(api.GroupLayer())
        if modifier.lottie.name:
            layer.desc = modifier.lottie.name

        inner = self._modifier_inner_group(modifier, shapegroup, layer)
        if isinstance(modifier.lottie, objects.Repeater):
            self.build_repeater(modifier.lottie, inner, layer)

    def _build_repeater_defs(self, shape, name_id):
        dup = api.Duplicate()
        dup.id = name_id
        self.canvas.defs.append(dup)
        self.canvas.register_as(dup, name_id)

        def getter(keyframe):
            if keyframe is None:
                v = shape.copies.value
            else:
                v = keyframe.start[0]

            return v - 1

        setattr(dup, "from", self.process_vector_ext(shape.copies.keyframes, getter))
        dup.to.value = 0
        dup.step.value = -1
        return dup

    def _build_repeater_transform_scale_component(self, shape, name_id, comp, scalecomposite):
        power = ast.SifPower()
        setattr(scalecomposite, "xy"[comp], power)

        def getter(keyframe):
            if keyframe is None:
                v = shape.transform.scale.value
            else:
                v = keyframe.start
            v = v[comp] / 100
            return v

        power.base = self.process_vector_ext(shape.transform.scale.keyframes, getter)

        # HACK work around an issue in Synfig
        power.power = ast.SifAdd()
        power.power.lhs.value = api.ValueReference(name_id)
        power.power.rhs.value = 0.000001

    def _build_repeater_transform(self, shape, inner, name_id):
        offset_id = name_id + "_origin"
        origin = api.ExportedValue(offset_id, self.process_vector(shape.transform.anchor_point), "vector")
        self.canvas.defs.append(origin)
        self.canvas.register_as(origin, offset_id)
        inner.origin = origin

        composite = inner.transformation

        composite.offset = ast.SifAdd()
        composite.offset.rhs.value = api.ValueReference(offset_id)
        composite.offset.lhs = ast.SifScale()
        composite.offset.lhs.scalar.value = api.ValueReference(name_id)
        composite.offset.lhs.link = self.process_vector(shape.transform.position)

        composite.angle = ast.SifScale()
        composite.angle.scalar.value = api.ValueReference(name_id)
        composite.angle.link = self.process_scalar(shape.transform.rotation)

        composite.scale = ast.SifVectorComposite()
        self._build_repeater_transform_scale_component(shape, name_id, 0, composite.scale)
        self._build_repeater_transform_scale_component(shape, name_id, 1, composite.scale)

    def _build_repeater_amount(self, shape, inner, name_id):
        inner.amount = ast.SifSubtract()
        inner.amount.lhs = self.process_scalar(shape.transform.start_opacity, 0.01)

        inner.amount.rhs = ast.SifScale()
        inner.amount.rhs.scalar.value = api.ValueReference(name_id)

        def getter(keyframe):
            if keyframe is None:
                t = 0
                end = shape.transform.end_opacity.value
            else:
                t = keyframe.time
                end = keyframe.start[0]
            start = shape.transform.start_opacity.get_value(t)
            n = shape.copies.get_value(t)
            v = (start - end) / (n - 1) / 100 if n > 0 else 0
            return v
        inner.amount.rhs.link = self.process_vector_ext(shape.transform.end_opacity.keyframes, getter)

    def build_repeater(self, shape, inner, dom_parent):
        name_id = "duplicate_%s" % next(self.autoid)
        dup = self._build_repeater_defs(shape, name_id)
        self._build_repeater_transform(shape, inner, name_id)
        self._build_repeater_amount(shape, inner, name_id)
        inner.desc = "Transformation for " + (dom_parent.desc or "duplicate")

        # duplicate layer
        duplicate = dom_parent.add_layer(api.DuplicateLayer())
        duplicate.index = dup
        duplicate.desc = shape.name


def to_sif(animation):
    builder = SifBuilder()
    builder.process(animation)
    return builder.canvas
